import openmm.openmm as mm
from .utils import CONST
from .unit import *


def slab_correction(system):
    '''
    Apply Yeh's long range coulomb correction for slab geometry in z direction
    to eliminate the undesired interactions between periodic slabs.

    It's useful for 2-D systems simulated under 3-D periodic condition.
    For this correction to work correctly:

    * A vacuum space two times larger than slab thickness is required.
    * All particles should never diffuse across the z boundaries.
    * The box size should not change during the simulation.

    Parameters
    ----------
    system : mm.System
        The OpenMM system to be simulated

    Returns
    -------
    force : mm.CustomCVForce
    '''
    muz = mm.CustomExternalForce('q*z')
    muz.addPerParticleParameter('q')
    nbforce = [f for f in system.getForces() if f.__class__ == mm.NonbondedForce][0]
    qsum = 0
    for i in range(nbforce.getNumParticles()):
        q = nbforce.getParticleParameters(i)[0].value_in_unit(qe)
        muz.addParticle(i, [q])
        qsum += q
    if abs(qsum) > 1E-4:
        raise Exception('Slab correction is not valid for non-neutral system')

    box = system.getDefaultPeriodicBoxVectors()
    vol = (box[0][0] * box[1][1] * box[2][2]).value_in_unit(nm ** 3)
    # convert from e^2/nm to kJ/mol  # 138.93545915168772
    _eps0 = CONST.EPS0 * farad / meter  # vacuum dielectric constant
    _conv = (1 / (4 * CONST.PI * _eps0) * qe ** 2 / nm / item).value_in_unit(kJ_mol)
    prefactor = 2 * CONST.PI / vol * _conv
    cvforce = mm.CustomCVForce(f'{prefactor}*muz*muz')
    cvforce.addCollectiveVariable('muz', muz)
    system.addForce(cvforce)

    return cvforce


def spring_self(system, positions, particles, strength):
    '''
    Restrain the selected particles at their original positions.

    Note that the original positions will NOT change with box size if there is a barostat.

    Parameters
    ----------
    system : mm.System
        The OpenMM system to be simulated
    positions : array_like
        The positions of all particles in the system in unit of nm
    particles : list of int
        The indexes of particles to be restrained
    strength : list of float
        The strength of harmonic restraint in x, y and z directions.
        Three elements should be provided in unit of kJ/mol/nm^2

    Returns
    -------
    force : mm.CustomExternalForce
    '''
    if system.getNumParticles() != len(positions):
        raise Exception('Length of positions does not equal to number of particles in system')

    kx, ky, kz = strength
    force = mm.CustomExternalForce(f'{kx}*periodicdistance(x,0,0,x0,0,0)^2+'
                                   f'{ky}*periodicdistance(0,y,0,0,y0,0)^2+'
                                   f'{kz}*periodicdistance(0,0,z,0,0,z0)^2')
    force.addPerParticleParameter('x0')
    force.addPerParticleParameter('y0')
    force.addPerParticleParameter('z0')
    for i in particles:
        force.addParticle(i, list(positions[i]))
    system.addForce(force)

    return force


def wall_power(system, particles, direction, bound, k, cutoff, power=2):
    '''
    Add a power wall for selected particles so that they cannot cross it.

    Note that periodic box condition is not considered,
    so you need to make sure particles will not move to other cells during the simulation.

    The energy equal to k when particle is located at the lower or higher bound,
    and equal to zero when particle is located between [lower bound + cutoff, higher bound - cutoff].

    Parameters
    ----------
    system : mm.System
    particles : list of int
    direction : ['x', 'y', 'z']
    bound : list of float
    k : float
    cutoff : float
    power : int, optional

    Returns
    -------
    force : mm.CustomExternalForce

    '''
    if direction not in ['x', 'y', 'z']:
        raise Exception('direction can only be x, y or z')
    _min, _max = bound
    _min_0 = _min + cutoff
    _max_0 = _max - cutoff
    force = mm.CustomExternalForce(f'{k}*step({_min_0}-{direction})*rmin^{power}+'
                                   f'{k}*step({direction}-{_max_0})*rmax^{power};'
                                   f'rmin=({_min_0}-{direction})/{cutoff};'
                                   f'rmax=({direction}-{_max_0})/{cutoff}')
    for i in particles:
        force.addParticle(i, [])
    system.addForce(force)

    return force


def wall_lj126(system, particles, direction, bound, epsilon, sigma):
    '''
    Add a LJ-12-6 wall for selected particles so that they cannot cross it.

    Note that periodic box condition is not considered,
    so you need to make sure particles will not move to other cells during the simulation.

    The energy is infinite when particle is located at the lower or higher bound,
    and equal to epsilon when particle is located at lower bound + sigma or higher bound - sigma,
    and equal to zero when particle is located between [lower bound + sigma * 2^(1/6), higher bound - sigma * 2^(1/6)].

    Parameters
    ----------
    system : mm.System
    particles : list of int
    direction : str
    bound : list of float
    epsilon : float
    sigma : float

    Returns
    -------
    force : mm.CustomExternalForce

    '''
    if direction not in ['x', 'y', 'z']:
        raise Exception('direction can only be x, y or z')
    _min, _max = bound
    _min_0 = _min + sigma * 2 ** (1 / 6)
    _max_0 = _max - sigma * 2 ** (1 / 6)
    force = mm.CustomExternalForce(f'4*{epsilon}*step({_min_0}-{direction})*(rmin^12-rmin^6+0.25)+'
                                   f'4*{epsilon}*step({direction}-{_max_0})*(rmax^12-rmax^6+0.25);'
                                   f'rmin={sigma}/({direction}-{_min});'
                                   f'rmax={sigma}/({_max}-{direction})')
    for i in particles:
        force.addParticle(i, [])
    system.addForce(force)

    return force


def electric_field(system, particles, strength):
    '''
    Apply external electric field to selected particles in a system.

    The unit of electric field strength is V/nm.

    Parameters
    ----------
    system : mm.System
    particles : list of int
    strength : list of float
        Strength of electric field in x, y and z directions, in unit of volt/nm

    Returns
    -------
    force : mm.CustomExternalForce

    '''
    # convert from eV/nm to kJ/(mol*nm)  # 96.4853400990037
    _conv = (1 * qe * volt / item).value_in_unit(kJ_mol)
    efx, efy, efz = strength
    force = mm.CustomExternalForce(f'{_conv}*({efx}*q*x+{efy}*q*y+{efz}*q*z)')
    force.addPerParticleParameter('q')
    nbforce = next(f for f in system.getForces() if type(f) is mm.NonbondedForce)
    for i in particles:
        q = nbforce.getParticleParameters(i)[0].value_in_unit(qe)
        force.addParticle(i, [q])
    system.addForce(force)

    return force


def CLPolCoulTT(system, donors, b=45.0):
    '''
    Apply Tang-Toennies damping for the Coulomb interactions between selected H-bond donors and Drude dipoles.

    Parameters
    ----------
    system : mm.System
    donors : list of int
        Indexes of particles served as H-bond donors
    b : float, optional
        b in unit of /nm

    Returns
    -------
    force : mm.CustomNonbondedForce

    '''
    nbforce: mm.NonbondedForce = next(f for f in system.getForces() if type(f) == mm.NonbondedForce)
    dforce: mm.DrudeForce = next(f for f in system.getForces() if type(f) == mm.DrudeForce)

    drude_pairs = {}  # {parent: drude}
    dipole_set = set()
    for i in range(dforce.getNumParticles()):
        drude, parent, p2, p3, p4, q, alpha, aniso12, aniso34 = dforce.getParticleParameters(i)
        drude_pairs[parent] = drude
        dipole_set.add(parent)
        dipole_set.add(drude)

    ttforce = mm.CustomNonbondedForce('-%.6f*q1*q2/r*beta*gamma;'
                                      'beta=exp(-br);'
                                      'gamma=1+br+br*br/2+br2*br/6+br2*br2/24;'
                                      'br2=br*br;'
                                      'br=%.6f*r' % (CONST.ONE_4PI_EPS0, b))
    ttforce.addPerParticleParameter('q')
    for i in range(system.getNumParticles()):
        if i in drude_pairs:
            q, _, _ = nbforce.getParticleParameters(drude_pairs[i])
            q = -q
        else:
            q, _, _ = nbforce.getParticleParameters(i)
        q = q.value_in_unit(qe)
        ttforce.addParticle([q])
    ttforce.setNonbondedMethod(mm.CustomNonbondedForce.CutoffPeriodic)
    ttforce.setCutoffDistance(1.2 * nm)
    ttforce.addInteractionGroup(set(donors), dipole_set)
    ttforce.setForceGroup(9)
    # map all the exclusions from NonbondedForce
    for i in range(nbforce.getNumExceptions()):
        ii, jj, _, _, _ = nbforce.getExceptionParameters(i)
        ttforce.addExclusion(ii, jj)
    system.addForce(ttforce)

    return ttforce


def restrain_particle_number(system, particles, direction, bound,
                             sigma, target, k, weights=None):
    '''
    Restrain the number of selected particles in a region.

    The region is defined by direction (x, y or z) and bound (lower and upper).
    Each particle is consider as a Gaussian distribution with standard deviation equal to sigma.
    The number of particles is restrained to the target value
    using a harmonic function with force constant k.

    Parameters
    ----------
    system : mm.System
    particles : list of int
    direction : ['x', 'y', 'z']
    bound : list of float
    sigma : float
        Variance of the particle Gaussian in unit of nm
    target : float
    k : float
        Strength of the harmonic restraint in unit of kJ/mol
    weights : list of float, optional

    Returns
    -------
    force : mm.CustomCVForce

    '''
    if direction not in ['x', 'y', 'z']:
        raise Exception('direction can only be x, y or z')
    _min, _max = bound
    if weights is None:
        weights = [1.0] * len(particles)
    if len(weights) != len(particles):
        raise Exception('particles and weights should have the same length')

    if _min is not None:
        str_min = f'erf(({_min}-{direction})/{2 ** 0.5 * sigma})'
    else:
        str_min = '-1'

    if _max is not None:
        str_max = f'erf(({_max}-{direction})/{2 ** 0.5 * sigma})'
    else:
        str_max = '1'

    nforce = mm.CustomExternalForce(f'0.5*({str_max}-{str_min})*weight')
    nforce.addPerParticleParameter('weight')
    for i, w in zip(particles, weights):
        nforce.addParticle(i, [w])

    cvforce = mm.CustomCVForce(f'0.5*{k}*(number-{target})^2')
    cvforce.addCollectiveVariable('number', nforce)
    system.addForce(cvforce)

    return cvforce


def point_wall_power(system, particles, reference, k, cutoff, power=2):
    '''
    Add a spherical wall centered at a reference point so that selected particles are repelled from this point.

    The energy equal to k when particle is located at the reference point,
    and equal to zero when particle stays beyond the cutoff.

    Parameters
    ----------
    system : mm.System
    particles : list of int
    reference : list of float
    k : float
    cutoff : float
    power : int, optional

    Returns
    -------
    force : mm.CustomExternalForce
    '''
    x0, y0, z0 = reference
    force = mm.CustomExternalForce(f'step({cutoff}-r)*{k}*(1-r/{cutoff})^{power};'
                                   f'r=periodicdistance(x,y,z,{x0},{y0},{z0});')
    for i in particles:
        force.addParticle(i, [])
    force.setName('PointWall')
    system.addForce(force)

    return force



class FFTermNotFoundError(Exception):
    pass


class TypingNotSupportedError(Exception):
    pass


class TypingUndefinedError(Exception):
    pass


class ChargeIncrementNonZeroError(Exception):
    pass



import os
import subprocess
from subprocess import Popen, PIPE
from pathlib import Path
from mstk import logger
from .pbsjob import PbsJob
from .scheduler import Scheduler


class Slurm(Scheduler):
    '''
    Slurm job scheduler with support of GPU allocation and MPI/OpenMP hybrid parallelization.

    Slurm is a rather powerful and *complicated* job scheduler with tons of configurations and options.
    It is not the goal of `mstk` to provided a comprehensive wrapper for Slurm.
    Therefore, it's very likely that the job script generated by this class doesn't fully fit the requirement of a specific computing center.
    In that case, it's viable to do some process on the generated job script before submitting it.

    Parameters
    ----------
    queue : str
        The jobs will be submitted to this partition.
    n_proc : int
        The CPU cores a job can use.
    n_gpu : int
        The GPU card a job can use.
    n_node : int
        The nodes a job can use. If 0, then will be decided by slurm.
    exclude: str, Optional
        The nodes to be excluded, in Slurm format
    env_cmd : str, Optional
        The commands for setting up the environment before running real calculations.
        It will be inserted on the top of job scripts.

    Attributes
    ----------
    queue : str
        The jobs will be submitted on this queue.
    n_proc : int
        The CPU cores a job can use.
    n_gpu : int
        The GPU card a job can use.
    n_node : int
        The nodes a job can use. If 0, then will be decided by slurm.
    exclude: str
        The nodes to be excluded, in Slurm format
    env_cmd : str
        The commands for setting up the environment before running real calculations.
    sh : str
        The default name of the job script
    max_running_hour: int
        The wall time limit for a job in hours.
    username : str
        The current user
    cached_jobs_expire : int
        The lifetime of cached jobs in seconds.
    submit_cmd : str
        The command for submitting the job script.
        If is `sbatch` by default. But extra argument can be provided, e.g. `sbatch --qos=debug`.
    '''

    #: Whether or not this is a remote job scheduler
    is_remote = False

    def __init__(self, queue, n_proc, n_gpu, n_node=0, exclude=None, env_cmd=None):
        super().__init__(queue=queue, n_proc=n_proc, n_gpu=n_gpu, env_cmd=env_cmd)
        self.n_node = n_node
        self.exclude = exclude or ''
        self.sh = '_job_slurm.sh'
        self.submit_cmd = 'sbatch'

    def is_working(self) -> bool:
        '''
        Check whether or not Slurm is working normally on this machine.

        It calls `sinfo --version` and check the output.

        Returns
        -------
        is : bool
        '''
        cmd = 'sinfo --version'
        sp = Popen(cmd.split(), stdout=PIPE, stderr=PIPE)
        stdout, stderr = sp.communicate()
        return stdout.decode().startswith('slurm')

    def generate_sh(self, commands, name, workdir=None, sh=None, id_prior=None, **kwargs):
        '''
        Generate a shell script for commands to be executed by the job scheduler on compute nodes.

        Because of the complexity of Slurm configurations, it's probable that the job script generated here is not fully valid.
        In that case, it's viable to do some process on the generated job script before submitting it.

        Parameters
        ----------
        commands : list of str
            The commands to be executed step by step
        name : str
            The name of the job to be submitted
        workdir : str, Optional
            The working directory of this job
        id_prior: int, Optional
            The id of prior job this one depends on
        sh : str, Optional
            The name (path) of shell script to be generated.
            If not set, will use the default :attr:`sh`.
        '''
        workdir = workdir or (self.remote_dir if self.is_remote else os.getcwd())
        workdir = Path(workdir).absolute().as_posix()
        sh = sh or self.sh
        sh_basename = Path(sh).stem
        out = sh_basename + '.out'
        err = sh_basename + '.err'
        node_cmd = f'#SBATCH --nodes={self.n_node}\n' if self.n_node > 0 else ''
        gpu_cmd = f'#SBATCH --gres=gpu:{self.n_gpu}\n' if self.n_gpu > 0 else ''
        dep_cmd = f'#SBATCH --dependency=afterok:{id_prior}\n' if id_prior is not None else ''  # id_prior can be 0
        exclude_cmd = f'#SBATCH --exclude={self.exclude}\n' if self.exclude else ''

        with open(sh, 'w') as f:
            f.write(f'#!/bin/bash\n'
                    f'#SBATCH -J {name}\n'
                    f'#SBATCH -D {workdir}\n'
                    f'#SBATCH -o {out}\n'
                    f'#SBATCH -e {err}\n'
                    f'{dep_cmd}'
                    f'#SBATCH -p {self.queue}\n'
                    f'#SBATCH --time={self.max_running_hour}:00:00\n'
                    f'{node_cmd}'
                    f'#SBATCH --ntasks={self.n_proc}\n'
                    f'{gpu_cmd}'
                    f'{exclude_cmd}'
                    f'\n'
                    f'{self.env_cmd}\n\n'
                    )
            for cmd in commands:
                f.write(cmd + '\n')

    def submit(self, sh=None):
        sh = sh or self.sh
        cmd = self.submit_cmd + ' ' + sh
        p = Popen(cmd.split(), stdout=PIPE, stderr=PIPE)
        b_out, b_err = p.communicate()
        out, err = b_out.strip().decode(), b_err.strip().decode()
        if p.returncode == 0 and out.startswith('Submitted batch job'):
            logger.info(f'{out}')
            return int(out.split()[-1])
        else:
            if out:
                logger.error(f'{out}')
            if err:
                logger.error(f'{err}')
            if not out and not err:
                logger.error(f'sbatch failed')
            return -1

    def kill_job(self, name) -> bool:
        job = self.get_job_from_name(name)
        if job is None:
            return False

        cmd = f'scancel {job.id}'
        return subprocess.call(cmd.split()) == 0

    def get_all_jobs(self):
        # Show all jobs. Then check the user
        cmd = 'scontrol show job'
        sp = Popen(cmd.split(), stdout=PIPE, stderr=PIPE)
        stdout, stderr = sp.communicate()
        if sp.returncode != 0:
            logger.error(stderr.decode())
            return []

        jobs = []
        for job_str in stdout.decode().split('\n\n'):  # split jobs
            if job_str.startswith('JobId'):
                job = self._get_job_from_str(job_str)
                # Show all jobs. Then check the user
                if job.user == self.username:
                    jobs.append(job)
        return jobs

    def _get_job_from_str(self, job_str) -> PbsJob:
        workdir = None
        for line in job_str.split():  # split properties
            try:
                key, val = line.split('=')[0:2]
            except:
                continue
            if key == 'JobId':
                id = int(val)
            elif key == 'UserId':
                user = val.split('(')[0]  # UserId=username(uid)
            elif key == 'JobName' or key == 'Name':
                name = val
            elif key == 'Partition':
                queue = val
            elif key == 'JobState':
                state_str = val
                if val in ('PENDING', 'RESV_DEL_HOLD'):
                    state = PbsJob.State.PENDING
                elif val in ('CONFIGURING', 'RUNNING', 'COMPLETING', 'STOPPED', 'SUSPENDED'):
                    state = PbsJob.State.RUNNING
                else:
                    state = PbsJob.State.DONE
            elif key == 'WorkDir':
                workdir = val
        job = PbsJob(id=id, name=name, state=state, workdir=workdir, user=user, queue=queue)
        job.state_str = state_str
        return job



import numpy as np
import openmm
from logging import Logger
from openmm import openmm as mm, app
from .grofile import GroFile
from .unit import kelvin, bar, nm, ps


class CONST:
    PI = 3.1415926535
    EPS0 = 8.8541878128E-12  # F/m
    ONE_4PI_EPS0 = 138.935455  # qq/r -> kJ/mol


def print_omm_info(logger=None):
    messages = ['OpenMM verison: ' + openmm.version.version,
                'OpenMM library: ' + openmm.version.openmm_library_path,
                'OpenMM platforms: ' + ', '.join([mm.Platform.getPlatform(i).getName()
                                                  for i in range(mm.Platform.getNumPlatforms())]),
                ]
    for msg in messages:
        if logger:
            logger.info(msg)
        else:
            print(msg)


def minimize(sim, tolerance, gro_out=None, logger=None):
    '''
    Run energy minimization on a Simulation until the force on each atom is smaller than tolerance.

    Note that the tolerance here differs from that in OpenMM LocalEnergyMinimizer.
    In OpenMM, the tolerance is set for the root-mean-square of N*3 force components in the whole system.

    Parameters
    ----------
    sim : app.Simulation
    tolerance : float
    gro_out : str
    logger : Logger
    '''
    if logger:
        logger.info(f'Initial energy = {sim.context.getState(getEnergy=True).getPotentialEnergy()}')
    i1_iter = 0
    tol_iter = tolerance
    while True:
        i1_iter += 1
        sim.minimizeEnergy(tolerance=tol_iter)
        forces = sim.context.getState(getForces=True).getForces(asNumpy=True)._value
        fsq = np.sum(forces * forces, axis=1)
        imax = np.argmax(fsq)
        fmax = fsq[imax] ** 0.5
        if logger:
            logger.info(f'Iter {i1_iter} Max force F_{imax} = {fmax} kJ/mol/nm')
        if fmax < tolerance or i1_iter >= 10:
            break
        tol_iter /= 2

    state = sim.context.getState(getPositions=True, getEnergy=True)
    if logger:
        logger.info(f'Minimized energy = {state.getPotentialEnergy()}')

    if gro_out:
        GroFile.writeFile(sim.topology, state.getPositions(), state.getPeriodicBoxVectors(), gro_out)


def apply_mc_barostat(system, pcoupl, P, T, nstep=100, rigid_molecule=True, logger=None):
    '''
    Add a MonteCarlo barostat to the system and return the added barostat

    TODO It requires the forked OpenMM https://github.com/openmm/openmm/pull/3797

    Parameters
    ----------
    system : mm.System
    pcoupl : str
        The type of barostat. Can be one of the following: iso, semi-iso, aniso, xy, z.
    P : float
        Pressure in bar
    T : float
        Temperature in Kelvin
    nstep : int
    rigid_molecule : bool
    logger : Logger

    Returns
    -------
    barostat : mm.Force
        The added barostat force
    '''
    if pcoupl == 'iso':
        msg = 'Isotropic barostat'
        force = mm.MonteCarloBarostat(P * bar, T * kelvin, nstep, rigid_molecule)
    elif pcoupl == 'semi-iso':
        msg = 'Anisotropic barostat with coupled XY'
        force = mm.MonteCarloMembraneBarostat(P * bar, 0 * bar * nm, T * kelvin,
                                              mm.MonteCarloMembraneBarostat.XYIsotropic,
                                              mm.MonteCarloMembraneBarostat.ZFree,
                                              nstep, rigid_molecule)
    elif pcoupl == 'aniso':
        msg = 'Anisotropic barostat'
        force = mm.MonteCarloAnisotropicBarostat([P * bar] * 3, T * kelvin, True, True, True, nstep, rigid_molecule)
    elif pcoupl == 'xy':
        msg = 'Anisotropic barostat only for X and Y'
        force = mm.MonteCarloAnisotropicBarostat([P * bar] * 3, T * kelvin, True, True, False, nstep, rigid_molecule)
    elif pcoupl == 'z':
        msg = 'Anisotropic barostat only for Z'
        force = mm.MonteCarloAnisotropicBarostat([P * bar] * 3, T * kelvin, False, False, True, nstep, rigid_molecule)
    else:
        raise Exception('Available pressure coupling types: iso, semi-iso, aniso, xy, z')

    scaled_unit = 'molecules' if rigid_molecule else 'constrained groups'
    msg += f' with {scaled_unit} scaled'

    if logger:
        logger.info(msg)

    system.addForce(force)

    return force


def energy_decomposition(sim: app.Simulation, logger=None):
    '''
    Get the energy contribution of different force groups

    Parameters
    ----------
    sim : app.Simulation
    logger : Logger

    Returns
    -------
    energies : Dict[str, float]
        The name and energy contribution of each force group
    '''
    groups = {}  # {id_group: {force1_name, force2_name, ...}, ...}
    for force in sim.system.getForces():
        i = force.getForceGroup()
        if i not in groups:
            groups[i] = set()
        groups[i].add(force.getName())

    energies = {}
    for i in sorted(groups.keys()):
        energy = sim.context.getState(getEnergy=True, groups={i}).getPotentialEnergy()._value
        if energy != 0 or i != 0:
            name = f'{i}_{"+".join(groups[i])}'
            energies[name] = energy

    if logger:
        for k, v in energies.items():
            logger.info(f'E_{k} = {v}')

    return energies



from openmm import app
from ..unit import *


class ViscosityReporter(object):
    '''
    ViscosityReporter report the viscosity using cosine periodic perturbation method.
    A integrator supporting this method is required.
    e.g. the VVIntegrator from https://github.com/z-gong/openmm-velocityVerlet.

    Parameters
    ----------
    file : string
        The file to write to
    reportInterval : int
        The interval (in time steps) at which to write frames
    append : bool
        Whether or not append to the existing file.
    '''

    def __init__(self, file, reportInterval, append=False):
        self._reportInterval = reportInterval
        if append:
            self._out = open(file, 'a')
        else:
            self._out = open(file, 'w')
        self._hasInitialized = False

    def describeNextReport(self, simulation):
        """Get information about the next report this object will generate.

        Parameters
        ----------
        simulation : Simulation
            The Simulation to generate a report for

        Returns
        -------
        tuple
            A six element tuple. The first element is the number of steps
            until the next report. The next four elements specify whether
            that report will require positions, velocities, forces, and
            energies respectively.  The final element specifies whether
            positions should be wrapped to lie in a single periodic box.
        """
        try:
            simulation.integrator.getCosAcceleration()
        except AttributeError:
            raise Exception('This integrator does not calculate viscosity')

        steps = self._reportInterval - simulation.currentStep % self._reportInterval
        return (steps, False, False, False, False)

    def report(self, simulation: app.Simulation, state):
        """Generate a report.

        Parameters
        ----------
        simulation : Simulation
            The Simulation to generate a report for
        state : State
            The current state of the simulation
        """
        if not self._hasInitialized:
            self._hasInitialized = True
            print('#"Step"\t"Acceleration (nm/ps^2)"\t"VelocityAmplitude (nm/ps)"\t"1/Viscosity (1/Pa.s)"', file=self._out)

        acceleration = simulation.integrator.getCosAcceleration().value_in_unit(nm / ps ** 2)
        vMax, invVis = simulation.integrator.getViscosity()
        vMax = vMax.value_in_unit(nm / ps)
        invVis = invVis.value_in_unit((unit.pascal * unit.second) ** -1)
        print(simulation.currentStep, acceleration, vMax, invVis, sep='\t', file=self._out)

        if hasattr(self._out, 'flush') and callable(self._out.flush):
            self._out.flush()

    def __del__(self):
        self._out.close()



class Singleton():
    _instance = None
    def __new__(cls, *args, **kwargs):
        if not isinstance(cls._instance, cls):
            cls._instance = object.__new__(cls, *args, **kwargs)
        return cls._instance


import os
import logging
from pathlib import Path

DIR_MSTK = Path(__file__).parent.as_posix()

MSTK_FORCEFIELD_PATH = [os.path.join(DIR_MSTK, 'data', 'forcefield')]
_ff_path_str = os.getenv('MSTK_FORCEFIELD_PATH')
if _ff_path_str:
    MSTK_FORCEFIELD_PATH = _ff_path_str.strip(':').split(':') + MSTK_FORCEFIELD_PATH


class _CustomFormatter(logging.Formatter):
    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = "%(asctime)s %(levelname)s (%(filename)s:%(lineno)d) %(message)s"

    FORMATS = {
        logging.DEBUG   : grey + format + reset,
        logging.INFO    : grey + format + reset,
        logging.WARNING : yellow + format + reset,
        logging.ERROR   : red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, "%m-%d %H:%M:%S")
        return formatter.format(record)


def _get_logger():
    logger = logging.getLogger('mstk')
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler()
    handler.setFormatter(_CustomFormatter())
    logger.addHandler(handler)
    return logger


logger = _get_logger()



#!/usr/bin/env python3

import sys
import argparse
import random
import numpy as np
from mstk.topology import Topology, Molecule
from mstk.trajectory import Trajectory, Frame
from mstk.forcefield import ForceField
from mstk.forcefield.typer import ZftTyper, GaffTyper
from mstk.forcefield.errors import *
from mstk.chem import constant
from mstk import logger


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--top', nargs='+', type=str, required=True,
                        help='topology files for molecules. '
                             'String starts with : will be treated as SMILES')
    parser.add_argument('-c', '--conf', nargs='+', type=str,
                        help='configuration file for each molecule. The last frame will be used. '
                             'Positions are required to build box and write coordinates. '
                             'If the topology doesn\'t contain positions, it must be provided by this argument. '
                             'The number of conf files can be smaller or equal to the number of molecule types')
    parser.add_argument('--op', type=str, default='top.psf', help='output topology file')
    parser.add_argument('--oc', type=str, default='conf.gro', help='output configuration file')
    parser.add_argument('-n', '--number', nargs='+', type=int, help='number of each molecule')
    parser.add_argument('-f', '--ff', nargs='+', required=True, type=str, help='forcefield files')
    parser.add_argument('-t', '--typer', type=str,
                        help='typing file. Required if SMILES provided for topology. '
                             'If set to gaff, will use the GAFF typing engine')
    parser.add_argument('--ua', action='store_true',
                        help='use united-atom model. Remove non-polar hydrogens (connected to C/Si) from molecules')
    parser.add_argument('--box', nargs='+', type=float,
                        help='periodic box size. '
                             'The box will be rectangular if three values are provided. '
                             'The box will be cubic if only one value is provided.'
                        )
    parser.add_argument('--density', type=float,
                        help='the density of the target box in g/mL. '
                             'The size of cubic box will be calculated from it. '
                             'If the argument box is provided, density will be ignored')
    parser.add_argument('--packmol', action='store_true',
                        help='generate Packmol input files instead of write coordinates directly')
    return parser.parse_args()


def main():
    args = parse_args()
    for arg, val in vars(args).items():
        logger.info(f'--{arg:10s} {val}')

    if args.typer == 'gaff':
        typer = GaffTyper('gaff.zft')
    elif args.typer:
        typer = ZftTyper(args.typer)
    else:
        typer = None

    ref_mols = []
    for inp in args.top:
        if inp.startswith(':') or inp.lower().endswith('.smi'):
            if typer is None:
                raise Exception('--typer is required for SMILES input')
        ref_mols += Topology.open(inp).molecules

    args.number = args.number or [1] * len(ref_mols)
    if len(args.number) != len(ref_mols):
        raise Exception('Inconsistent molecules and numbers')

    ff = ForceField.open(*args.ff)

    for mol in ref_mols:
        logger.info(f'Processing {mol}...')
        mol.remove_drude_particles()
        mol.remove_virtual_sites()
        if mol.n_atom > 1 and mol.n_bond == 0:
            logger.warning(f'{mol} carries no bond. Make sure the topology is correct')

        if typer is not None:
            try:
                typer.type(mol)
            except TypingNotSupportedError as e:
                logger.warning(f'Failed typing {mol}: {e}. '
                               f'This warning can be ignored if atom type is already assigned')
            except TypingUndefinedError as e:
                xyz = '_typing_' + mol.name + '.xyz'
                Topology([mol]).write(xyz)
                logger.error(f'Failed typing {mol}: {e}. Check {xyz}')
                sys.exit(1)

        if args.ua:
            mol.remove_non_polar_hydrogens()

        if ff.is_polarizable:
            mol.generate_drude_particles(ff)
        if ff.has_virtual_site:
            mol.generate_virtual_sites(ff)
        ff.assign_mass(mol)
        ff.assign_charge(mol)

    conf_files = args.conf or []
    for mol, conf_file in zip(ref_mols, conf_files):
        frame = Trajectory.read_frame_from_file(conf_file, -1)
        _positions_set = False
        if len(frame.positions) == mol.n_atom:
            mol.set_positions(frame.positions)
            _positions_set = True
        elif ff.is_polarizable or ff.has_virtual_site:
            atoms_real = [atom for atom in mol.atoms if not atom.is_drude and atom.virtual_site is None]
            if len(frame.positions) == len(atoms_real):
                for i, atom in enumerate(atoms_real):
                    atom.position = frame.positions[i][:]
                np.random.seed(1)
                for parent, drude in mol.get_drude_pairs():
                    drude.position = parent.position + (np.random.random(3) - 0.5) / 100
                for parent, vsite in mol.get_virtual_site_pairs():
                    vsite.position = vsite.virtual_site.calc_position()
                _positions_set = True

        if not _positions_set:
            logger.error(f'Numbers of atoms in {conf_file} (%i) and {mol} (all: %i, real: %i) do not match' % (
                len(frame.positions), mol.n_atom,
                mol.n_atom - len(mol.get_drude_pairs()) - len(mol.get_virtual_site_pairs())))
            sys.exit(1)

    top = Topology(ref_mols, args.number)

    box = [0.0, 0.0, 0.0]
    if args.density is not None:
        mass = sum(atom.mass for atom in top.atoms)  # g/mol
        vol = mass / args.density / constant.AVOGADRO  # cm^3
        box = [vol ** (1 / 3) * constant.CENTI / constant.NANO] * 3  # nm

    if args.box is not None:
        if len(args.box) == 3:
            box = args.box
        elif len(args.box) == 1:
            box = args.box * 3
        else:
            raise Exception('box should be one or three floats')

    logger.info(f'box = {box}')
    top.cell.set_box(box)

    top.write(args.op)

    if args.packmol:
        top.update_molecules(ref_mols)
        # write the Packmol input files only. Can modify it before execution
        top.scale_with_packmol(args.number, seed=random.randrange(int(1E5), int(1E6)), tempdir='.')
    else:
        frame = Frame(top.n_atom)
        frame.cell = top.cell
        frame.positions = top.positions
        trj = Trajectory.open(args.oc, 'w')
        trj.write_frame(frame, top)
        trj.close()


if __name__ == '__main__':
    main()



from mstk.chem.element import Element
from mstk.topology.atom import Atom
from mstk.topology.molecule import Molecule
from mstk.topology.topology import Topology


class GroTopology:
    '''
    Generate Topology from GRO file.

    Atom name, residue information, unit cell and positions are parsed.
    The atom name colume will be used as both atom name and atom type.
    GRO file does not contain connectivity information. Each residue will be considered as a molecule.

    Only the first frame will be considered.

    Parameters
    ----------
    file : str

    Attributes
    ----------
    topology : Topology

    Examples
    --------
    >>> gro = Gro('input.gro')
    >>> topology = gro.topology

    >>> Gro.save_to(topology, 'output.gro')
    '''

    def __init__(self, file, **kwargs):
        self.topology = Topology()
        self._parse(file)

    def _parse(self, file):
        molecules = {}  # {int: Molecule}
        n_atom = 0
        last_resid = None
        with open(file) as f:
            for i, line in enumerate(f):
                if i == 0:
                    continue
                if i == 1:
                    n_atom = int(line)
                    continue
                if i == n_atom + 2:
                    _box = tuple(map(float, line.split()))
                    if len(_box) == 3:
                        self.topology.cell.set_box(_box)
                    elif len(_box) == 9:
                        ax, by, cz, ay, az, bx, bz, cx, cy = _box
                        self.topology.cell.set_box([[ax, ay, az], [bx, by, bz], [cx, cy, cz]])
                    else:
                        raise ValueError('Invalid box')
                    break

                res_id = int(line[:5])
                res_name = line[5:10].strip()
                atom_name = line[10:15].strip()
                element = Element.guess_from_atom_type(atom_name)
                atom_id = int(line[15:20])
                x = float(line[20:28])
                y = float(line[28:36])
                z = float(line[36:44])

                atom = Atom(atom_name)
                atom.type = atom_name
                atom.symbol = element.symbol
                atom.mass = element.mass
                atom.position = (x, y, z,)

                if res_id != last_resid:
                    mol = Molecule(res_name)
                    molecules[res_id] = mol
                    last_resid = res_id
                else:
                    mol = molecules[res_id]
                mol.add_atom(atom)

        self.topology.update_molecules(list(molecules.values()))

    @staticmethod
    def save_to(top, file, **kwargs):
        '''
        Save topology into a GRO file

        Parameters
        ----------
        top : Topology
        file : str
        '''
        if not top.has_position:
            raise Exception('Position is required for writing GRO file')

        def format_float(value, max_length, default_decimal):
            if value >= 10 ** max_length or value <= -10 ** (max_length - 1):
                raise Exception(f'{value} too long for formatter')
            formatter = f'%{max_length}.{default_decimal}f'
            return (formatter % value)[:max_length]

        string = 'Created by mstk\n'
        string += f'{top.n_atom}\n'

        for atom in top.atoms:
            pos = atom.position
            string += "%5i%-5s%5s%5i%8.3f%8.3f%8.3f\n" % (
                (atom.residue.id + 1) % 100000, atom.residue.name[:5], atom.name[:5], (atom.id + 1) % 100000,
                pos[0], pos[1], pos[2]
            )

        if top.cell.is_rectangular:
            string += ' %.4f %.4f %.4f\n' % tuple(top.cell.size)
        else:
            a, b, c = top.cell.vectors
            string += ' %.4f %.4f %.4f %.4f %.4f %.4f %.4f %.4f %.4f\n' % (
                a[0], b[1], c[2], a[1], a[2], b[0], b[2], c[0], c[1])

        with open(file, 'wb') as f:
            f.write(string.encode())


Topology.register_format('.gro', GroTopology)



import os
from .system import System
from ..forcefield import *
from ..topology import *
from .. import logger


class LammpsExporter():
    '''
    LammpsExporter export a non-polarizable :class:`System` to input files for LAMMPS.

    LAMMPS is powerful and flexible. But the input file for LAMMPS is a mess and cannot be handled in an elegant way,
    especially when there are hybrid functional forms.
    Here, only limited types of FF (OPLS-AA and SDK-CG) will be considered.

    The following potential functions are currently supported:

    * :class:`~mstk.forcefield.LJ126Term`
    * :class:`~mstk.forcefield.MieTerm`
    * :class:`~mstk.forcefield.HarmonicBondTerm`
    * :class:`~mstk.forcefield.HarmonicAngleTerm`
    * :class:`~mstk.forcefield.SDKAngleTerm`
    * :class:`~mstk.forcefield.OplsDihedralTerm`
    * :class:`~mstk.forcefield.PeriodicDihedralTerm`
    * :class:`~mstk.forcefield.OplsImproperTerm`

    In order to run simulation of SDK-CG model, the package CG-SDK should be included when compiling LAMMPS.
    '''

    def __init__(self):
        pass

    @staticmethod
    def export(system, data_out, in_out, **kwargs):
        '''
        Generate input files for LAMMPS from a system

        Parameters
        ----------
        system : System
        data_out : str
        in_out : str
        '''
        if not system.use_pbc:
            raise Exception('PBC required for exporting LAMMPS')

        supported_terms = {LJ126Term, MieTerm,
                           HarmonicBondTerm, MorseBondTerm,
                           HarmonicAngleTerm, SDKAngleTerm, LinearAngleTerm,
                           OplsDihedralTerm, PeriodicDihedralTerm,
                           OplsImproperTerm}
        unsupported = system.ff_classes - supported_terms
        if unsupported != set():
            raise Exception('Unsupported FF terms: %s' % (', '.join(map(lambda x: x.__name__, unsupported))))

        if system.topology.has_virtual_site:
            raise Exception('Virtual sites not supported by LAMMPS')

        if LinearAngleTerm in system.ff_classes:
            logger.warning('LinearAngleTerm not supported by LAMMPS. Exported in harmonic form at 178 degree')

        top = system.topology
        ff = system.ff

        # Morse bond requires hybrid bond styles
        is_morse = MorseBondTerm in system.ff_classes

        # SDK CGFF requires all pairwise parameters and hybrid angle styles
        is_sdk = MieTerm in system.ff_classes or SDKAngleTerm in system.ff_classes

        ### Assign LAMMPS types ####################################
        bond_types = list(ff.bond_terms.values())
        angle_types = list(ff.angle_terms.values())
        dihedral_types = list(ff.dihedral_terms.values())
        improper_types = list(ff.improper_terms.values())

        lmp_types: {str: Atom} = {}  # {'typ': Atom}
        lmp_types_atype: {str: AtomType} = {}  # {'typ': AtomType}

        for atom in top.atoms:
            if atom.type not in lmp_types:
                lmp_types[atom.type] = atom
                lmp_types_atype[atom.type] = ff.atom_types[atom.type]

        lmp_type_list = list(lmp_types.keys())
        lmp_symbol_list = [atom.symbol or 'UNK' for atom in lmp_types.values()]
        lmp_shake_bonds = [i + 1 for i, bterm in enumerate(bond_types) if bterm.fixed]
        lmp_shake_angles = [i + 1 for i, aterm in enumerate(angle_types) if aterm.fixed]
        #############################################################

        string = 'Created by mstk\n'

        string += '\n%i atoms\n' % top.n_atom
        string += '%i bonds\n' % top.n_bond
        string += '%i angles\n' % top.n_angle
        string += '%i dihedrals\n' % top.n_dihedral
        string += '%i impropers\n' % top.n_improper

        string += '\n%i atom types\n' % len(lmp_type_list)
        string += '%i bond types\n' % len(bond_types)
        string += '%i angle types\n' % len(angle_types)
        string += '%i dihedral types\n' % len(dihedral_types)
        string += '%i improper types\n' % len(improper_types)

        if not top.cell.is_rectangular:
            raise Exception('Triclinic box haven\'t been implemented')
        box = top.cell.size * 10
        string += '\n0 %10.4f  xlo xhi\n' % box[0]
        string += '0 %10.4f  ylo yhi\n' % box[1]
        string += '0 %10.4f  zlo zhi\n' % box[2]

        string += '\nMasses\n\n'

        for i, (typ, atom) in enumerate(lmp_types.items()):
            string += '%4i %8.3f  # %8s\n' % (i + 1, atom.mass, typ)

        if len(bond_types) > 0:
            string += '\nBond Coeffs  # %s\n\n' % ('hybrid harmonic morse' if is_morse else 'harmonic')

        for i, bterm in enumerate(bond_types):
            bond_style = '' if not is_morse else 'morse' if type(bterm) is MorseBondTerm else 'harmonic'
            if type(bterm) is HarmonicBondTerm:
                string += '%4i %8s %12.6f %10.4f  # %s-%s\n' % (
                    i + 1, bond_style, bterm.k / 4.184 / 100, bterm.length * 10, bterm.type1, bterm.type2)
            elif type(bterm) is MorseBondTerm:
                alpha = (bterm.k / bterm.depth) ** 0.5
                string += '%4i %8s %12.6f %10.4f %10.4f  # %s-%s\n' % (
                    i + 1, bond_style, bterm.depth / 4.184, alpha / 10, bterm.length * 10, bterm.type1, bterm.type2)

        if len(angle_types) > 0:
            string += '\nAngle Coeffs  # %s\n\n' % ('hybrid harmonic sdk' if is_sdk else 'harmonic')

        for i, aterm in enumerate(angle_types):
            angle_style = '' if not is_sdk else 'sdk' if type(aterm) is SDKAngleTerm else 'harmonic'
            string += '%4i %8s %12.6f %10.4f  # %s-%s-%s\n' % (
                i + 1, angle_style, aterm.k / 4.184, min(aterm.theta * RAD2DEG, 178),
                aterm.type1, aterm.type2, aterm.type3)

        if len(dihedral_types) > 0:
            string += '\nDihedral Coeffs  # opls\n\n'

        for i, dterm in enumerate(dihedral_types):
            if type(dterm) is PeriodicDihedralTerm:
                try:
                    dterm = dterm.to_opls_term()
                except:
                    raise Exception('Only OPLS type dihedral is implemented')
            k1, k2, k3, k4 = map(lambda x: x * 2 / 4.184, [dterm.k1, dterm.k2, dterm.k3, dterm.k4])
            string += '%4i %12.6f %12.6f %12.6f %12.6f  # %s-%s-%s-%s\n' % (
                i + 1, k1, k2, k3, k4,
                dterm.type1, dterm.type2, dterm.type3, dterm.type4)

        if len(improper_types) > 0:
            string += '\nImproper Coeffs  # cvff\n\n'

        for i, iterm in enumerate(improper_types):
            string += '%4i %12.6f %4i %4i  # %s-%s-%s-%s\n' % (
                i + 1, iterm.k / 4.184, -1, 2, iterm.type2, iterm.type3, iterm.type1, iterm.type4)

        string += '\nAtoms\n\n'

        for atom in top.atoms:
            typ = atom.type
            x, y, z = atom.position * 10
            string += '%8i %6i %6i %12.6f %10.4f %10.4f %10.4f  # %8s %8s\n' % (
                atom.id + 1, atom.molecule.id + 1, lmp_type_list.index(typ) + 1,
                atom.charge, x, y, z, atom.name, atom.molecule.name)

        if top.n_bond > 0:
            string += '\nBonds\n\n'

        for i, bond in enumerate(top.bonds):
            a1, a2 = bond.atom1, bond.atom2
            btype = bond_types.index(system.bond_terms[bond]) + 1
            string += '%6i %6i %6i %6i  # %s-%s\n' % (
                i + 1, btype, a1.id + 1, a2.id + 1, a1.name, a2.name)

        if top.n_angle > 0:
            string += '\nAngles\n\n'

        for i, angle in enumerate(top.angles):
            aterm = angle_types.index(system.angle_terms[angle]) + 1
            a1, a2, a3 = angle.atom1, angle.atom2, angle.atom3
            string += '%6i %6i %6i %6i %6i  # %s-%s-%s\n' % (
                i + 1, aterm, a1.id + 1, a2.id + 1, a3.id + 1, a1.name, a2.name, a3.name)

        if top.n_dihedral > 0:
            string += '\nDihedrals\n\n'

        for i, dihedral in enumerate(top.dihedrals):
            dtype = dihedral_types.index(system.dihedral_terms[dihedral]) + 1
            a1, a2, a3, a4 = dihedral.atom1, dihedral.atom2, dihedral.atom3, dihedral.atom4
            string += '%6i %6i %6i %6i %6i %6i  # %s-%s-%s-%s\n' % (
                i + 1, dtype, a1.id + 1, a2.id + 1, a3.id + 1, a4.id + 1,
                a1.name, a2.name, a3.name, a4.name)

        if top.n_improper > 0:
            string += '\nImpropers\n\n'

        for i, improper in enumerate(top.impropers):
            itype = improper_types.index(system.improper_terms[improper]) + 1
            a1, a2, a3, a4 = improper.atom1, improper.atom2, improper.atom3, improper.atom4
            string += '%6i %6i %6i %6i %6i %6i  # %s-%s-%s-%s\n' % (
                i + 1, itype, a2.id + 1, a3.id + 1, a1.id + 1, a4.id + 1,
                a2.name, a3.name, a1.name, a4.name)

        with open(data_out, 'wb')  as f:
            f.write(string.encode())

        ### LAMMPS input script #################################################
        cmd_bond = 'hybrid harmonic morse' if is_morse else 'harmonic'
        cmd_angle = 'hybrid harmonic sdk' if is_sdk else 'harmonic'
        cmd_pair_base = 'lj/sdk' if is_sdk else 'lj/cut'
        cmd_pair_suffix = '/coul/long' if system.charged else ''
        cmd_mix = 'arithmetic' if system.ff.lj_mixing_rule == ForceField.LJ_MIXING_LB else 'geometric'
        cmd_tail = 'yes' if ff.vdw_long_range == ForceField.VDW_LONGRANGE_CORRECT else 'no'
        cmd_shift = 'yes' if ff.vdw_long_range == ForceField.VDW_LONGRANGE_SHIFT else 'no'
        cmd_comment_kspace = '# ' if not system.charged else ''

        cmd_paras = ''
        for i, (typ_i, atom) in enumerate(lmp_types.items()):
            for j, (typ_j, atom) in enumerate(lmp_types.items()):
                if j < i:
                    continue
                atype1 = lmp_types_atype[typ_i]
                atype2 = lmp_types_atype[typ_j]
                try:
                    # LAMMPS CG-SDK does not allow lj mixing. have to write it explicitly
                    vdw = ff.get_vdw_term(atype1, atype2, mixing=is_sdk)
                except:
                    continue

                if is_sdk:
                    if type(vdw) is MieTerm:
                        if vdw.is_sdk:
                            lj_style = f'lj{int(vdw.repulsion)}_{int(vdw.attraction)}'
                        else:
                            raise Exception('MieTerm other than SDK type are not implemented')
                    else:
                        lj_style = 'lj12_6'
                    cmd_paras += 'pair_coeff %4i %4i %8s %9.5f %8.4f  # %8s %8s %s\n' % (
                        i + 1, j + 1, lj_style, vdw.epsilon / 4.184, vdw.sigma * 10,
                        lmp_type_list[i], lmp_type_list[j], ','.join(vdw.comments))
                else:
                    cmd_paras += 'pair_coeff %4i %4i %9.5f %8.4f  # %8s %8s %s\n' % (
                        i + 1, j + 1, vdw.epsilon / 4.184, vdw.sigma * 10,
                        lmp_type_list[i], lmp_type_list[j], ','.join(vdw.comments))

        cmd_shake = 'fix SHAKE all shake 0.0001 20 0 b '
        if lmp_shake_bonds:
            cmd_shake += ' '.join(map(str, lmp_shake_bonds))
            if lmp_shake_angles:
                cmd_shake += ' a ' + ' '.join(map(str, lmp_shake_angles))
        else:
            cmd_shake = '# ' + cmd_shake
        timestep = 10.0 if is_sdk else 1.0

        string = f'''# created by mstk
units real
boundary p p p
atom_style full
bond_style {cmd_bond}
angle_style {cmd_angle}
dihedral_style opls
improper_style cvff
special_bonds lj 0 0 {ff.scale_14_vdw} coul 0 0 {ff.scale_14_coulomb}

pair_style {cmd_pair_base}{cmd_pair_suffix} {ff.vdw_cutoff * 10}
pair_modify mix {cmd_mix} tail {cmd_tail} shift {cmd_shift}
{cmd_comment_kspace}kspace_style pppm 1.0e-4

read_data {os.path.basename(data_out)}

{cmd_paras}

variable T equal 300
variable P equal 1
variable elec equal ecoul+elong

# thermo_style custom step press pe ebond eangle edihed eimp evdwl v_elec
# thermo 10
# minimize 1.0e-4 1.0e-6 200 1000
# reset_timestep 0

{cmd_shake}
fix ICECUBE all momentum 100 linear 1 1 1

velocity all create $T 12345
timestep {timestep}

fix NPT all npt temp $T $T {timestep * 100} iso $P $P {timestep * 500} fixedpoint 0 0 0

thermo_style custom step temp press pe emol evdwl v_elec density
thermo_modify flush yes
thermo 1000

variable slog equal logfreq(10,9,10)
dump TRJ all custom 10 dump.lammpstrj id mol type element q xu yu zu
dump_modify TRJ sort id element {' '.join(lmp_symbol_list)} every v_slog first yes
dump DCD all dcd 10000 dump.dcd
dump_modify DCD unwrap yes

restart 1000000 rst_*
run 1000000
'''

        with open(in_out, 'wb') as f:
            f.write(string.encode())



import os
import openmm.openmm as mm


class CheckpointReporter():
    '''
    CheckpointReporter saves periodic checkpoints of a simulation.
    The checkpoints will overwrite old files -- only the latest three will be kept.
    State XML files can be saved together, in case the checkpoint files are broken.

    Parameters
    ----------
    file : string
        The file to write to.
        Any current contents will be overwritten.
        The latest three checkpoint will be kept with the step appended to the file name.
    reportInterval : int
        The interval (in time steps) at which to write checkpoints.
    xml : string, optional
        If provided, the state will be serialized into XML format and saved together with checkpoint.
        Any current contents will be overwritten.
        The latest three XML files will be kept with the step appended to the file name.
    '''

    def __init__(self, file, reportInterval, xml=None):
        self._reportInterval = reportInterval
        self._file = file
        self._xml = xml

        if type(file) is not str:
            raise Exception('file should be str')

    def describeNextReport(self, simulation):
        """Get information about the next report this object will generate.

        Parameters
        ----------
        simulation : Simulation
            The Simulation to generate a report for

        Returns
        -------
        tuple
            A five element tuple. The first element is the number of steps
            until the next report. The remaining elements specify whether
            that report will require positions, velocities, forces, and
            energies respectively.
        """
        steps = self._reportInterval - simulation.currentStep % self._reportInterval
        return (steps, True, True, False, False, False)

    def report(self, simulation, state):
        """Generate a report.

        Parameters
        ----------
        simulation : Simulation
            The Simulation to generate a report for
        state : State
            The current state of the simulation
        """

        filename = self._file + '_%i' % simulation.currentStep
        with open(filename, 'wb') as out:
            out.write(simulation.context.createCheckpoint())

        file_prev3 = self._file + '_%i' % (simulation.currentStep - 3 * self._reportInterval)
        if os.path.exists(file_prev3):
            os.remove(file_prev3)

        if self._xml is not None:
            xml_name = self._xml + '_%i' % simulation.currentStep
            xml = mm.XmlSerializer.serialize(state)
            with open(xml_name, 'w') as f:
                f.write(xml)

            xml_prev3 = self._xml + '_%i' % (simulation.currentStep - 3 * self._reportInterval)
            if os.path.exists(xml_prev3):
                os.remove(xml_prev3)



import io
import os
from .typer import Typer
from mstk import DIR_MSTK, MSTK_FORCEFIELD_PATH
from mstk.forcefield.errors import *
from mstk.topology import Bond

try:
    from rdkit.Chem import AllChem as Chem
except:
    RDKIT_IMPORTER = False
else:
    RDKIT_IMPORTER = True


class TypeDefine():
    def __init__(self, name, smarts, formal_charge=None):
        self.name = name
        self.smarts = smarts
        self.rdsmarts = None
        self.formal_charge = formal_charge  # None means don't match formal charge
        # smarts not provided means this is a fake root define
        if smarts is not None:
            try:
                self.rdsmarts = Chem.MolFromSmarts(smarts)
            except:
                raise Exception('Invalid SMARTS: %s' % smarts)

        self.children: [TypeDefine] = []
        self.parent: TypeDefine = None

    def __repr__(self):
        return '<TypeDefine: %s %s %s>' % (self.name, self.smarts, self.formal_charge)

    def add_child(self, define):
        if define not in self.children:
            self.children.append(define)
            define.parent = self


class ZftTyper(Typer):
    '''
    ZftTyper assign atom types with local environment defined by SMARTS and a hierarchical rule.

    A type definition file is required.

    Examples
    --------
    >>> TypeDefinition

    >>> H_1 [#1]
    >>> C_4 [#6X4]
    >>> HC  [H][CX4]
    >>> CT  [CX4;H3]
    >>> CS  [CX4;H2]
    >>> HT  [H][CX4]c1ccccc1


    >>> HierarchicalTree

    >>> H_1
    >>>     HC
    >>>         HT
    >>> C_4
    >>>     CT
    >>>     CS

    This is a sample type definition file for hydrocarbons with OPLS force field.
    Two sections are required: `TypeDefinition` and `HierarchicalTree`.
    `TypeDefinition` determines which type the atom can be by matching SMARTS.
    A atom can match several different atom types defined in `TypeDefinition` section.
    e.g. Hydrogen atoms with one neighbour will match type `H_1`, and carbon atoms with four neighbour will match type `C_4`.
    Hydrogen atoms bonded with four neighboured carbon will also match type `HC`.
    If a four neighboured carbon is bonded to a benzene ring, then hydrogen atoms bonded to this carbon will also match type `HT`.
    Four neighboured carbon connected with two and three hydrogen atoms will also match `CS` and `CT`, respectively.

    After all the possible atom types are matched,
    `HierarchicalTree` determines which type the atom will finally be by performing depth first search.
    The indentation (by 4 spaces) in the `HierarchicalTree` section represents the depth of atom types.
    Therefore, if a atom matches `C_4`, but not `CT` or `CS`, it will be typed as `C_4`
    If a atom matches both `C_4` and `CT` but not `CS`, it will be typed as `CT`.
    If a atom matches both `C_4` and `CS` but not `CT`, it will be typed as `CS`.
    If a atom matches both `C_4`, `CT` and `CS` (which is impossible in this example), it will be typed as `CT`.

    For molecule propane, based on the `TypeDefinition`, all the hydrogen atoms will match atom type `H_1` and `HC`.
    All of the carbon atoms will match atom type `C_4`, but the side carbon atoms will also match `CT` and the center carbon atom will also match `CS`.
    Then based on the `HierarchicalTree`, all of the hydrogen atoms will be typed as `HC`.
    Side carbon atoms will be typed as `CT`, and center carbon atom will be typed as `CS`.

    The hierarchical strategy make the atom type definition extendable.

    Parameters
    ----------
    file : str or file-like object, optional
        Type definition file.
        If the file does not exist, will search it under directories defined by `MSTK_FORCEFIELD_PATH`.

    Notes
    -----
    * SMARTS is parsed by using RDKit package. Make sure it is installed.
    * In type definition file, empty lines are ignored, and comments should start with ##.

    '''

    def __init__(self, file=None):
        if not RDKIT_IMPORTER:
            raise Exception('RDKit is required for ZftTyper')

        self.defines: {str: TypeDefine} = {}
        self.define_root = TypeDefine('UNDEFINED', None)

        content = None
        if type(file) is str:
            for dir in ['.'] + MSTK_FORCEFIELD_PATH:
                p = os.path.join(dir, file)
                if os.path.exists(p):
                    with open(p) as f:
                        content = f.read()
                    break
            else:
                raise Exception(f'Typing file not found: {file}')
        elif isinstance(file, io.IOBase):
            content = file.read()
        if content is None:
            raise Exception('Argument file or string should be provided')

        self._parse(content)

    def _parse(self, content):
        lines = content.splitlines()

        section = ''
        tree_lines = []
        for line in lines:
            line = line.split('##')[0]
            if line.strip() == '':
                continue
            if line.strip() == 'TypeDefinition':
                section = 'TypeDefinition'
                continue
            if line.strip() == 'HierarchicalTree':
                section = 'HierarchicalTree'
                continue
            if section == 'TypeDefinition':
                words = line.strip().split()
                if len(words) < 2:
                    raise Exception('Atom type and SMARTS must be provided: ' + line)
                name, smarts = words[:2]
                try:
                    formal_charge = int(words[2]) if len(words) > 2 else None
                except ValueError:
                    raise Exception('Formal charge must be integer: ' + line)
                self.defines[name] = TypeDefine(name, smarts, formal_charge)
            if section == 'HierarchicalTree':
                tree_lines.append(line.rstrip())

        last_level = 0
        last_define = self.define_root
        for line in tree_lines:
            name = line.lstrip()
            indent = len(line) - len(name)
            if indent % 4 != 0:
                raise Exception('Indentation for HierarchicalTree should be 4 spaces: %s' % line)
            level = (indent) // 4 + 1
            if level == last_level + 1:
                parent = last_define
            elif level <= last_level:
                parent = last_define
                for i in range(last_level - level + 1):
                    parent = parent.parent
            else:
                raise Exception('Invalid indentation: %s' % line)
            last_level = level
            try:
                last_define = self.defines[name]
            except:
                raise Exception('Atom type not found in TypeDefinition section: %s' % line)
            parent.add_child(last_define)

    def _type_molecule(self, molecule):
        '''
        Assign atom types in the molecule with rules in type definition file.

        The :attr:`~mstk.topology.Atom.type` attribute of all atoms in the molecule will be updated.

        ZftTyper use RDKit to do SMARTS matching, therefore the orders must be set for all the bonds in the molecule.
        Usually it means the molecule be initialized from SMILES.
        with :func:`~mstk.topology.Molecule.from_smiles` or :func:`~mstk.topology.Molecule.from_pybel`

        If an atom can not match any type by the predefined SMARTS patterns, an Exception will be raised.

        Parameters
        ----------
        molecule : Molecule
        '''
        if any(bond.order == Bond.Order.UNSPECIFIED for bond in molecule.bonds):
            raise TypingNotSupportedError(f'Typer requires all bond orders in {molecule} be set')

        possible_defines = {i: [] for i in range(molecule.n_atom)}
        for define in self.defines.values():
            for indexes in molecule.rdmol.GetSubstructMatches(define.rdsmarts, uniquify=False, maxMatches=0):
                if define.formal_charge is not None \
                        and sum(molecule.atoms[i].formal_charge for i in indexes) != define.formal_charge:
                    continue
                idx = indexes[0]
                possible_defines[idx].append(define)

        _undefined = []
        for i, atom in enumerate(molecule.atoms):
            define = self._get_deepest_define(possible_defines[i], self.define_root)
            if define is self.define_root:
                _undefined.append(atom.name)
            else:
                atom.type = define.name
        if _undefined != []:
            raise TypingUndefinedError('Definition not found for %i atoms: %s' % (
                len(_undefined), ' '.join(_undefined)))

    def _get_deepest_define(self, defines, parent: TypeDefine):
        for define in parent.children:
            if define in defines:
                return self._get_deepest_define(defines, define)
        return parent



#!/usr/bin/env python3

import sys
import argparse
import math
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from mstk.topology import Topology
from mstk.trajectory import Trajectory
from mstk.utils import print_data_to_file
from mstk.topology.geometry import periodic_distance
from mstk import logger

np.seterr(all='raise')
matplotlib.use('Agg')
matplotlib.rcParams.update({'font.size': 15})


def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('input', nargs='+', type=str,
                        help='trajectory files for atomic positions and charges')
    parser.add_argument('-p', '--top', required=True, type=str,
                        help='psf or lammps data file for topology information')
    parser.add_argument('--a1', type=str, nargs='+', help='type of central atoms')
    parser.add_argument('--a2', type=str, nargs='+', help='type of coordinate atoms')
    parser.add_argument('--m1', type=str, help='name of central molecules')
    parser.add_argument('--m2', type=str, help='name of coordinate molecules')
    parser.add_argument('-o', '--output', required=True, type=str, help='Output prefix')
    parser.add_argument('-b', '--begin', default=0, type=int,
                        help='first frame to analyze. Index starts from 0')
    parser.add_argument('-e', '--end', default=-1, type=int,
                        help='last frame (not included) to analyze. Index starts from 0. '
                             '-1 means until last frames (included)')
    parser.add_argument('--maxr', default=1.0, type=float,
                        help='max distance (nm) for calculation of distribution')
    parser.add_argument('--dr', default=0.01, type=float,
                        help='bin size (nm) for calculation of distribution')
    parser.add_argument('--ignore', nargs='+', default=[], type=str,
                        help='ignore these molecules in topology in case topology and trajectory do not match')
    parser.add_argument('--skip', default=1, type=int, help='skip frames in trajectory')

    return parser.parse_args()


if __name__ == '__main__':
    args = parse_args()
    top = Topology.open(args.top)
    if args.ignore != []:
        molecules = [mol for mol in top.molecules if mol.name not in args.ignore]
        top.update_molecules(molecules)
    logger.info(top)

    trj = Trajectory.open(args.input)
    logger.info(trj)

    if (top.n_atom != trj.n_atom):
        raise Exception('Number of atoms in topology and trajectory files do not match')

    if args.m1 is None:
        group1 = [[atom] for atom in top.atoms if atom.type in args.a1]
    else:
        group1 = [[atom for atom in mol.atoms if atom.type in args.a1]
                  for mol in top.molecules if mol.name == args.m1]
    ids_group1 = [[atom.id for atom in atoms] for atoms in group1]
    masses_group1 = [[atom.mass for atom in atoms] for atoms in group1]

    if args.m2 is None:
        group2 = [[atom] for atom in top.atoms if atom.type in args.a2]
    else:
        group2 = [[atom for atom in mol.atoms if atom.type in args.a2]
                  for mol in top.molecules if mol.name == args.m2]
    ids_group2 = [[atom.id for atom in atoms] for atoms in group2]
    masses_group2 = [[atom.mass for atom in atoms] for atoms in group2]

    if args.end > trj.n_frame or args.end == -1:
        args.end = trj.n_frame

    dr = args.dr
    n_bin = math.ceil(args.maxr / dr) + 1
    edges = np.array([dr * (i - 0.5) for i in range(n_bin + 1)], dtype=float)
    r_array = (edges[1:] + edges[:-1]) / 2
    rdf_array = np.zeros(n_bin, dtype=float)


    def _get_weighted_center(positions, weight):
        return np.sum(positions * np.array(weight)[:, np.newaxis], axis=0) / sum(weight)


    def _get_com_group(positions, ids_group, masses_group):
        com_group = []
        for ids, masses in zip(ids_group, masses_group):
            poss = positions[ids]
            if len(ids) == 1:
                pos = poss[0]
            else:
                pos = _get_weighted_center(poss, masses)
            com_group.append(pos)
        return com_group


    n_frame = 0
    for i in range(args.begin, args.end, args.skip):
        n_frame += 1
        frame = trj.read_frame(i)
        sys.stdout.write('\r    frame %i' % i)

        com_group1 = _get_com_group(frame.positions, ids_group1, masses_group1)
        com_group2 = _get_com_group(frame.positions, ids_group2, masses_group2)

        vol = frame.cell.volume
        density_pair = len(group1) * len(group2) / vol

        density_array = np.zeros(len(r_array), dtype=float)
        for com1 in com_group1:
            for com2 in com_group2:
                distance = periodic_distance(com1, com2, frame.cell.size, args.maxr + dr / 2)
                if distance is not None and distance < args.maxr + dr / 2:
                    idx_r = int((distance - edges[0]) / dr)
                    density_array[idx_r] += 1

        rdf_array[1:] += density_array[1:] / (4 * np.pi * r_array[1:] ** 2 * dr) / density_pair

    rdf_array = rdf_array / n_frame

    name_column_dict = {'r'  : r_array,
                        'rdf': rdf_array}
    print_data_to_file(name_column_dict, f'{args.output}-rdf.txt')

    fig, ax = plt.subplots()
    ax.set(xlabel='r (nm)', ylabel='RDF')
    ax.plot(r_array, rdf_array, label='RDF')
    ax.legend()
    fig.tight_layout()
    fig.savefig(f'{args.output}-rdf.png')



#!/usr/bin/env python3

import argparse
from mstk.topology import Topology
from mstk.trajectory import Trajectory
from mstk.forcefield import ForceField
from mstk import logger


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('input', nargs='+', type=str, help='topology file')
    parser.add_argument('-n', '--number', nargs='+', type=int, help='number of molecules')
    parser.add_argument('-c', '--conf', type=str,
                        help='configuration file with positions and box. The last frame will be used. '
                             'This is for writing topology in formats supporting positions, e.g. PDB, XYZ')
    parser.add_argument('-o', '--output', required=True, type=str, help='output topology file')
    parser.add_argument('--ignore', nargs='+', default=[], type=str, help='ignore these molecule types')
    parser.add_argument('-f', '--ff', type=str, help='reassign charge from FF')
    parser.add_argument('--qscale', default=1, type=float, help='scale the charge of atoms')
    parser.add_argument('--qscaleignore', nargs='+', default=[], type=str,
                        help='ignore these molecule names for charge scaling')
    parser.add_argument('--box', nargs='+', type=float,
                        help='overwrite the box dimensions')
    parser.add_argument('--shift', nargs=3, default=[0, 0, 0], type=float,
                        help='shift the positions of all atoms')
    parser.add_argument('--ua', action='store_true',
                        help='remove hydrogen atoms bonded to C/Si. '
                             'Also remove the relevant bonds/angles/dihedrals/impropers')

    return parser.parse_args()


if __name__ == '__main__':
    args = parse_args()

    top_list = [Topology.open(inp) for inp in args.input]
    if args.number is None:
        args.number = [1] * len(top_list)

    molecules = []
    for top, n in zip(top_list, args.number):
        molecules.extend(top.molecules * n)
    if args.ignore:
        molecules = [mol for mol in molecules if mol.name not in args.ignore]

    top = top_list[0]
    top.update_molecules(molecules)
    logger.info(str(top))

    if args.conf:
        frame = Trajectory.read_frame_from_file(args.conf, -1)
        top.cell.set_box(frame.cell.vectors)
        top.set_positions(frame.positions)

    if args.ua:
        for mol in top.molecules:
            mol.remove_non_polar_hydrogens(update_topology=False)
        top.update_molecules(top.molecules)
        logger.info(str(top))

    if args.ff:
        ff = ForceField.open(args.ff)
        ff.assign_charge(top)
        logger.info(f'Net charge = {sum(a.charge for a in top.atoms)}')

    if args.qscale != 1:
        for mol in top.molecules:
            if mol.name in args.qscaleignore:
                continue
            for atom in mol.atoms:
                atom.charge *= args.qscale

    if args.box is not None:
        if len(args.box) == 3:
            box = args.box
        elif len(args.box) == 1:
            box = args.box * 3
        else:
            raise Exception('box should be float or list of three floats')
        top.cell.set_box(box)

    if top.has_position and set(args.shift) != {0}:
        for atom in top.atoms:
            atom.position += args.shift

    top.write(args.output, atom_type=True)



import numpy as np
from mstk import logger
from mstk.topology import UnitCell


class Frame:
    '''
    A frame records the positions and unit cell (and optionally velocities, charges, current step and current time)
    of the system at one specific time step.

    Velocity is stored for analyzing dynamic properties with auto correlation function.
    Charge is stored for fluctuating charge method.

    Parameters
    ----------
    n_atom : int
        The number of atoms in the frame.

    Attributes
    ----------
    step : int
        The current step. -1 means unknown.
    time : float
        The current simulation time. -1 means unknown.
    cell : UnitCell
        The unit cell.
    positions : array_like
        The positions of all atoms in the frame.
        It's a np.ndarry of shape (n_atom, 3)
    has_velocity : bool
        Whether or not there's velocity information for all atoms
    velocities : array_like
        The velocities of all atoms in the frame.
        It's a np.ndarry of shape (n_atom, 3)
    has_charge : bool
        Whether or not there's charge information for all atoms
    charges : array_like
        The charge of all atoms in the frame.
        It's a np.ndarry of shape (n_atom,)
    '''

    def __init__(self, n_atom):
        self.step = -1  # -1 means step information not found
        self.time = -1  # in ps. -1 means time information not found
        self.cell = UnitCell()
        self.n_atom = n_atom
        self.has_velocity = False
        self.has_charge = False
        self._positions = np.zeros((n_atom, 3), dtype=float)
        self._velocities = np.zeros((n_atom, 3), dtype=float)
        self._charges = np.zeros(n_atom, dtype=float)  # for fluctuating charge simulations

    def reset(self):
        '''
        Remove step, time and cell information, and reset has_velocity, has_charge to False.
        '''
        self.step = -1
        self.time = -1
        self.cell.set_box(None)
        self.has_velocity = False
        self.has_charge = False

    def resize(self, n_atom):
        '''
        Resize the number of atoms recorded in the frame.

        Usually it is not necessary to call this method explicitly.
        If n_atom is smaller than previous value, the information of the end atoms will be trimmed.
        If n_atom is larger than previous value, the information for the new atoms will be random.

        Parameters
        ----------
        n_atom : int
            The new number of atoms
        '''
        self.n_atom = n_atom
        self._positions.resize((n_atom, 3), refcheck=False)
        self._velocities.resize((n_atom, 3), refcheck=False)
        self._charges.resize((n_atom, 3), refcheck=False)

    @property
    def positions(self):
        '''
        The positions of all atoms in this frame

        :setter: Set the positions of all atoms

        Returns
        -------
        positions : array_like
            The positions is a numpy array of shape (n_atom, 3)
        '''
        return self._positions

    @positions.setter
    def positions(self, value):
        if not isinstance(value, (list, tuple, np.ndarray)) or len(value) != self.n_atom \
                or not isinstance(value[0], (list, tuple, np.ndarray)) or len(value[0]) != 3:
            raise ValueError(f'positions should have the shape of ({self.n_atom}, 3)')
        self._positions[:] = value

    @property
    def velocities(self):
        '''
        The velocities of all atoms in this frame

        :setter: Set the velocities of all atoms

        Returns
        -------
        velocities : array_like
            The velocities is a numpy array of shape (n_atom, 3)
        '''
        return self._velocities

    @velocities.setter
    def velocities(self, value):
        if not isinstance(value, (list, tuple, np.ndarray)) or len(value) != self.n_atom \
                or not isinstance(value[0], (list, tuple, np.ndarray)) or len(value[0]) != 3:
            raise ValueError(f'velocities should have the shape of ({self.n_atom}, 3)')
        self._velocities[:] = value

    @property
    def charges(self):
        '''
        The charges of all atoms in this frame

        :setter: Set the charges of all atoms

        Returns
        -------
        charges : array_like
            The charges is a numpy array of shape (n_atom,)
        '''
        return self._charges

    @charges.setter
    def charges(self, value):
        if not isinstance(value, (list, tuple, np.ndarray)) or len(value) != self.n_atom:
            raise ValueError(f'charges should have {self.n_atom} elements')
        self._charges[:] = value



class PbsJob():
    '''
    A PbsJob object is a job managed by job scheduler.

    Parameters
    ----------
    id : int
        Job ID assigned by job scheduler.
    name : str
        Name of this job in job scheduler.
    state: [0, 1, 9]
        Whether this job is pending (0), running (1) or finished (9).
        Killed or failed is also considered as finished.
    workdir : str
        The working directory of this job.
    user : str
        Name of the user who submitted this job.
    queue : str
        The queue this job is submitted to.

    Attributes
    ----------
    id : int
        Job ID assigned by job scheduler.
    name : str
        Name of this job in job scheduler.
    state: [0, 1, 9]
        Whether this job is pending (0), running (1) or finished (9).
        Killed or failed is also considered as finished.
    workdir : str
        The working directory of this job.
    user : str
        Name of the user who submitted this job.
    queue : str
        The queue this job is submitted to.
    state_str : str
        The string shown by the job scheduler, which represents the state of the job.
    '''

    def __init__(self, id, name, state, workdir=None, user=None, queue=None):
        self.id = id
        self.name = name
        self.state = state
        self.workdir = workdir
        self.user = user
        self.queue = queue
        self.state_str = None

    def __repr__(self):
        return '<PbsJob: %i %s %i %s>' % (self.id, self.name, self.state, self.user)

    def __eq__(self, other):
        return self.id == other.id

    class State:
        PENDING = 0
        RUNNING = 1
        DONE = 9



import numpy as np
from mstk.topology.atom import Atom
from mstk.topology.molecule import Molecule
from mstk.topology.topology import Topology
from mstk.chem.element import Element
from mstk import logger


class LammpsData():
    '''
    Generate topology from Lammps data file.

    Atom masses, types, charges, connectivity, unit cell and atom positions are parsed.
    The velocities and force field parameters are ignored.

    The string of atom types are assumed to be the comment after the mass.
    If such comment is not provided, then the string of atom types will be 'AT1', 'AT2'...
    If any atom type has mass smaller than 1.0 and atom type comment started with 'D',
    then it will be treated as Drude particle. e.g.

    >>> Masses
    >>>   1 1.008 # HC
    >>>   2 16.00 # OH
    >>>   3 0.800 # D

    HC and OH will be treated as the atom type string of the first and second atom types.
    The third atom type will be treated as Drude particle.

    Care should be taken for the impropers.
    The central atom for impropers in mstk is the first atom.
    But in Lammps data file, the central atom can be the first, second or third atom,
    depending on the `improper_style`.
    The argument `improper_center` controls which atom should be treated as the center atom,
    which is set to 1 by default.

    Parameters
    ----------
    file : str
    improper_center : [1, 2, 3, 4]
    kwargs : dict
    Ignored

    Examples
    --------

    >>> lammps = LammpsData('data.lmp')
    >>> topology = lammps.topology

    For `improper_style cvff`, the central atom is the third.
    Then the LammpsData should be loaded as

    >>> lammps = LammpsData('data.lmp', improper_center=3)

    Attributes
    ----------
    topology : Topology
        Parsed topology from the data
    n_atom_type : int
        Number of atom types
    type_names : list of str
        Names of these atom types
    type_masses : list of float
        Masses of these atom types
    xlo : float
        The origin of the cell in x direction. Not really useful
    ylo : float
        The origin of the cell in z direction. Not really useful
    zlo : float
        The origin of the cell in z direction. Not really useful
    '''

    def __init__(self, file, improper_center=None, **kwargs):
        self.topology = Topology()
        self._parse(file, improper_center)

    def _parse(self, file, improper_center):
        with open(file) as f:
            lines = f.read().splitlines()

        self.topology.remark = lines[0].strip()

        _sections = ["Atoms", "Velocities", "Ellipsoids", "Lines", "Triangles", "Bodies",
                     "Bonds", "Angles", "Dihedrals", "Impropers",
                     "Masses", "Pair Coeffs", "PairIJ Coeffs", "Bond Coeffs", "Angle Coeffs",
                     "Dihedral Coeffs", "Improper Coeffs",
                     "BondBond Coeffs", "BondAngle Coeffs", "MiddleBondTorsion Coeffs",
                     "EndBondTorsion Coeffs", "AngleTorsion Coeffs",
                     "AngleAngleTorsion Coeffs", "BondBond13 Coeffs", "AngleAngle Coeffs"]

        # record the line number (starts from 0) of section headers
        _iline_section = []
        for i, line in enumerate(lines):
            line_clean = line.split('#')[0].strip()
            if line_clean in _sections:
                _iline_section.append((i, line_clean))

        n_atom, n_bond, n_angle, n_dihedral, n_improper = \
            self._parse_header(lines[:_iline_section[0][0]])

        self.type_masses = [0.] * (self.n_atom_type + 1)
        self.type_names = [''] * (self.n_atom_type + 1)

        for i, (iline, section) in enumerate(_iline_section):
            if i != len(_iline_section) - 1:
                iline_next = _iline_section[i + 1][0]
            else:
                iline_next = len(lines)
            if section == 'Masses':
                # there is a blank line after each section header
                self._parse_atom_types(lines[iline + 2: iline_next])
            if section == 'Atoms':
                self._parse_atoms(lines[iline + 2: iline_next], n_atom)
            if section == 'Bonds':
                self._parse_bonds(lines[iline + 2: iline_next], n_bond)
            if section == 'Angles':
                self._parse_angles(lines[iline + 2: iline_next], n_angle)
            if section == 'Dihedrals':
                self._parse_dihedrals(lines[iline + 2: iline_next], n_dihedral)
            if section == 'Impropers':
                self._parse_impropers(lines[iline + 2: iline_next], n_improper, improper_center)

    def _parse_header(self, lines):
        n_atom = n_bond = n_angle = n_dihedral = n_improper = self.n_atom_type = 0
        la = lb = lc = bx = cx = cy = 0
        for line in lines:
            line = line.strip()
            words = line.split()
            if line.endswith(' atoms'):
                n_atom = int(words[0])
            elif line.endswith(' bonds'):
                n_bond = int(words[0])
            elif line.endswith(' angles'):
                n_angle = int(words[0])
            elif line.endswith(' dihedrals'):
                n_dihedral = int(words[0])
            elif line.endswith(' impropers'):
                n_improper = int(words[0])
            elif line.endswith(' atom types'):
                self.n_atom_type = int(words[0])
            elif line.endswith(' xlo xhi'):
                la = (float(words[1]) - float(words[0])) / 10
                self.xlo = float(words[0]) / 10
            elif line.endswith(' ylo yhi'):
                lb = (float(words[1]) - float(words[0])) / 10
                self.ylo = float(words[0]) / 10
            elif line.endswith(' zlo zhi'):
                lc = (float(words[1]) - float(words[0])) / 10
                self.zlo = float(words[0]) / 10
            elif line.endswith(' xy xz yz'):
                bx, cx, cy = tuple(map(lambda x: float(x) / 10, words[:3]))
            else:
                continue

        self.topology.cell.set_box([[la, 0, 0], [bx, lb, 0], [cx, cy, lc]])

        return n_atom, n_bond, n_angle, n_dihedral, n_improper

    def _parse_atom_types(self, lines):
        for i in range(self.n_atom_type):
            words = lines[i].strip().split()
            type_id = int(words[0])
            mass = float(words[1])
            self.type_masses[type_id] = mass
            # TODO This is not robust but there's no better way
            if len(words) > 3 and words[2] == '#':
                if words[-1].startswith('D') and mass < 1:
                    self.type_names[type_id] = 'DP_'
                    logger.warning(f'Atom type {type_id} is considered to be Drude particle')
                else:
                    self.type_names[type_id] = words[3]
            else:
                self.type_names[type_id] = 'AT' + str(type_id)

    def _parse_atoms(self, lines, n_atom):
        atoms = [Atom() for i in range(n_atom)]
        mol_names: {int: str} = {}
        for i in range(n_atom):
            words = lines[i].strip().split()
            atom_id = int(words[0])
            mol_id = int(words[1])
            type_id = int(words[2])
            charge = float(words[3])
            x = float(words[4]) / 10  # convert from A to nm
            y = float(words[5]) / 10
            z = float(words[6]) / 10
            try:
                ix = int(words[7])
                iy = int(words[8])
                iz = int(words[9])
            except:
                ix = iy = iz = 0

            if mol_id not in mol_names:
                mol_names[mol_id] = words[9] if (
                        len(words) > 9 and words[7] == '#') else 'M%i' % mol_id

            atom = atoms[atom_id - 1]
            atom._mol_id = mol_id  # temporary attribute for identifying molecules
            atom.id = atom_id - 1  # atom.id starts from 0
            atom.charge = charge
            atom.mass = self.type_masses[type_id]
            atom.type = self.type_names[type_id]
            if atom.type == 'DP_':
                atom.is_drude = True
                atom.symbol = 'DP'
            else:
                atom.symbol = Element.guess_from_atom_type(atom.type).symbol
            atom.position = np.array([x - self.xlo + ix * self.topology.cell.size[0],
                                      y - self.ylo + iy * self.topology.cell.size[1],
                                      z - self.zlo + iz * self.topology.cell.size[2]])

        molecules = [Molecule(name) for id, name in sorted(mol_names.items())]
        for atom in sorted(atoms, key=lambda x: x.id):
            mol = molecules[atom._mol_id - 1]
            mol.add_atom(atom, update_topology=False)
        for mol in molecules:
            for i, atom in enumerate(mol.atoms):
                # atomic symbol + index inside mol starting from 1
                atom.name = atom.symbol + str(i + 1)

        self.topology.update_molecules(molecules)

    def _parse_bonds(self, lines, n_bond):
        atoms = self.topology.atoms
        for i in range(n_bond):
            words = lines[i].strip().split()
            atom1 = atoms[int(words[2]) - 1]
            atom2 = atoms[int(words[3]) - 1]
            atom1.molecule.add_bond(atom1, atom2)

    def _parse_angles(self, lines, n_angle):
        atoms = self.topology.atoms
        for i in range(n_angle):
            words = lines[i].strip().split()
            atom1 = atoms[int(words[2]) - 1]
            atom2 = atoms[int(words[3]) - 1]
            atom3 = atoms[int(words[4]) - 1]
            atom1.molecule.add_angle(atom1, atom2, atom3)

    def _parse_dihedrals(self, lines, n_dihedral):
        atoms = self.topology.atoms
        for i in range(n_dihedral):
            words = lines[i].strip().split()
            atom1 = atoms[int(words[2]) - 1]
            atom2 = atoms[int(words[3]) - 1]
            atom3 = atoms[int(words[4]) - 1]
            atom4 = atoms[int(words[5]) - 1]
            atom1.molecule.add_dihedral(atom1, atom2, atom3, atom4)

    def _parse_impropers(self, lines, n_improper, improper_center):
        if improper_center is None:
            improper_center = 1
            logger.warning('improper_center undefined, '
                           'will treat the first atom as the central atom for impropers')
        elif improper_center not in (1, 2, 3, 4):
            raise Exception('improper_center should be a integer from 1 to 4')

        atoms = self.topology.atoms
        for i in range(n_improper):
            words = lines[i].strip().split()
            sides = []
            sides.append(atoms[int(words[2]) - 1])
            sides.append(atoms[int(words[3]) - 1])
            sides.append(atoms[int(words[4]) - 1])
            sides.append(atoms[int(words[5]) - 1])
            center = sides.pop(improper_center - 1)
            sides[0].molecule.add_improper(center, *sides)


Topology.register_format('.lmp', LammpsData)



import copy
import itertools
from ..topology import *
from ..forcefield import *
from .. import logger


class System:
    '''
    System is a combination of topology and forcefield, ready for simulation.

    When a system being initialized, a compressed set of force field is constructed with the terms required by the topology.
    Note that the terms in the compressed force field are references to the terms in the original force field.
    A series of data structures are also constructed to map the topological elements to force field terms.
    Unless you are certain, DO NOT make any modification to the topology and force field after a system is initialized, as it may corrupt the system.

    During the initialization, all information in the topology will be kept untouched, which means:

    * The atom type of all atoms in the topology must have been correctly assigned,
      so that the force field terms can be matched.
    * The charges and masses of all atoms in the topology must have been assigned.
      You may need to call :func:`mstk.forcefield.ForceField.assign_charge`
      and :func:`mstk.forcefield.ForceField.assign_mass`
      before constructing the system to update the charges and masses of atoms in the topology.
    * If this is a Drude polarizable model, the Drude particles must be available in the topology,
      and the polarizability, thole screening and charges on Drude particles must have been correctly assigned.
      You may need to call :func:`mstk.topology.Topology.generate_drude_particles` to generate Drude particles.
    * If this is a virtual site model, the virtual site particles must be available in the topology.

    The provided force field must be able to fully describe the energy of the topology.
    If any atom type from topology is not found in the FF, an Exception will be raised.
    If any vdw/bond/angle/dihedral/improper/polarizable term required by topology is not found in the FF,
    an Exception will be raised unless `allow_missing_terms` set to True.
    It won't check charge increment terms and virtual site terms, because they are already represented by the topology itself.

    Positions and unit cell should be included in the topology.
    If positions are not included in the topology, an Exception will be raised.
    Unit cell is optional. If unit cell is not provided, cutoff will not be used for non-bonded interactions.

    If some bonded (bond, angle, dihedral, improper) parameters required by the topology are not found in the force field,
    `mstk` can try to transfer parameters from more general terms in the force field.
    e.g. if bond term between 'c_4o2' and 'n_3' is not found, the bond term between 'c_4' and 'n_3' will be used.
    This mechanism is initially designed for TEAM force field.
    It is disabled by default, and can be enabled by setting argument `transfer_bonded_terms` to True.

    Parameters
    ----------
    topology : Topology
    ff : ForceField
    positions : array_like, optional
    cell : UnitCell, optional
    ignore_missing_improper: bool
        If set to True, a zero-energy OplsImproperTerm will be created for missing improper terms in the FF
    transfer_bonded_terms: bool
    suppress_pbc_warning: bool
        Set to True if you intend to build a vacuum system and don't want to hear warning about it.
    allow_missing_terms: bool
        If set to True, won't raise Exception if FF terms other than AtomType are missing.
        This is useful if you want to check which FF terms are missing.

    Attributes
    ----------
    charged : bool
        True if any atom in this system carries charge
    use_pbc : bool
        True if unit cell information provided and volume is not zero
    drude_pairs : dict, [Atom, Atom]
    vsite_pairs : dict, [Atom, Atom]
    constrain_bonds : dict, [Bond, float]
    constrain_angles : dict, [Angle, float]
    bond_terms : dict, [Bond, subclass of BondTerm]
    angle_terms : dict, [Angle, subclass of AngleTerm]
    dihedral_terms : dict, [Dihedral, subclass of DihedralTerm]
    improper_terms : dict, [Improper, subclass of ImproperTerm]
    polarizable_terms :  dict, [Atom, subclass of PolarizableTerm]
    vdw_classes : set of subclass of VdwTerm
    bond_classes : set of subclass of BondTerm
    angle_classes : set of subclass of AngleTerm
    dihedral_classes : set of subclass of DihedralTerm
    improper_classes : set of subclass of ImproperTerm
    polarizable_classes : set of subclass of PolarizableTerm
    ff_classes : set of subclass of FFTerm
    vsite_types : set of subclass of VirtualSite
    missing_terms : list of FFTerm
    '''

    def __init__(self, topology, ff,
                 ignore_missing_improper=False,
                 transfer_bonded_terms=False,
                 suppress_pbc_warning=False,
                 allow_missing_terms=False):
        self._topology = topology
        self._ff = ForceField()

        if not topology.has_position:
            raise Exception('Positions should be provided with topology')

        if self._topology.n_atom == 0:
            logger.error('No atoms in the topology')
            raise Exception('Invalid topology')

        self.use_pbc = bool(self._topology.cell.volume != 0)  # convert numpy.bool_ to bool
        if not self.use_pbc and not suppress_pbc_warning:
            logger.warning('Periodic cell not provided or volume equal to zero')

        if all(atom.mass <= 0 for atom in topology.atoms):
            logger.error('All atoms have non-positive mass. '
                         'You may need to call ff.assign_mass(topology)')
            raise Exception('Invalid topology')

        self.charged = True
        if all(atom.charge == 0 for atom in topology.atoms):
            self.charged = False
            logger.warning('All atoms have zero charge. '
                           'You may need to call ff.assign_charge(topology)')

        charge_total = sum(atom.charge for atom in topology.atoms)
        if abs(charge_total) > 1E-8:
            logger.warning('System is not charge neutral. Charge = %.6f' % charge_total)

        # prebuild topology information
        self.drude_pairs: {Atom: Atom} = dict(self._topology.get_drude_pairs())  # {parent: atom_drude}
        self.vsite_pairs: {Atom: Atom} = dict(self._topology.get_virtual_site_pairs())  # {parent: atom_vsite}
        # records the types of VirtualSite in topology. It is NOT the VirtualSiteTerm in force field
        self.vsite_types = set([type(atom.virtual_site) for atom in self.vsite_pairs.values()])

        # map topological elements to FF terms
        self.bond_terms: {Bond: BondTerm} = {}  # Drude bonds are not included
        self.angle_terms: {Angle: AngleTerm} = {}
        self.dihedral_terms: {Dihedral: DihedralTerm} = {}
        self.improper_terms: {Improper: ImproperTerm} = {}
        self.polarizable_terms: {Atom: PolarizableTerm} = {}  # key is parent Atom
        self.constrain_bonds: {Bond: float} = {}  # value is distance
        self.constrain_angles: {Angle: float} = {}  # value is 1-3 distance

        # missing FF terms
        self.missing_terms = []

        # unique FF term classes
        self.vdw_classes = set()
        self.bond_classes = set()
        self.angle_classes = set()
        self.dihedral_classes = set()
        self.improper_classes = set()
        self.polarizable_classes = set()
        self.ff_classes = set()

        self._extract_terms(ff, ignore_missing_improper, transfer_bonded_terms, allow_missing_terms)

    @property
    def topology(self):
        '''
        The topology associated with the system

        Returns
        -------
        topology : Topology
        '''
        return self._topology

    @property
    def ff(self):
        '''
        The compressed force field associated with the system

        Returns
        -------
        ff : ForceField
        '''
        return self._ff

    def _extract_terms(self, ff, ignore_missing_improper=False, transfer_bonded_terms=False, allow_missing_terms=False):
        '''
        Extract only the required terms from a force field, and record the missing terms.

        Parameters
        ----------
        ff : ForceField
        ignore_missing_improper : bool
        transfer_bonded_terms : bool
        allow_missing_terms : bool
        '''
        from mstk.forcefield.dff_utils import dff_fuzzy_match

        self._ff.restore_settings(ff.get_settings())

        _atype_not_found = set()
        for atom in self._topology.atoms:
            try:
                atype = ff.atom_types[atom.type]
            except:
                _atype_not_found.add(atom.type)
            else:
                self._ff.add_term(atype, replace=True)
        if _atype_not_found:
            logger.error('%i atom types not found in FF: %s' % (len(_atype_not_found), ' '.join(_atype_not_found)))
            raise Exception('Incomplete force field')

        _vdw_not_found = {}
        for atype in self._ff.atom_types.values():
            try:
                vdw = ff.get_vdw_term(atype, atype)
            except Exception as e:
                _vdw = VdwTerm(atype.eqt_vdw, atype.eqt_vdw)
                _vdw_not_found[_vdw.name] = _vdw
            else:
                self._ff.add_term(vdw, replace=True)

        for atype1, atype2 in itertools.combinations(self._ff.atom_types.values(), 2):
            try:
                vdw = ff.get_vdw_term(atype1, atype2, mixing=False)
            except:
                if self._ff.lj_mixing_rule == self._ff.LJ_MIXING_NONE:
                    # all pairwise vdW terms are explicitly required
                    _vdw = VdwTerm(atype1.eqt_vdw, atype2.eqt_vdw)
                    _vdw_not_found[_vdw.name] = _vdw
            else:
                self._ff.add_term(vdw, replace=True)

        _bterm_not_found = {}
        _bterm_transferred = set()
        for bond in self._topology.bonds:
            if bond.is_drude:
                # the Drude-parent bonds will be handled by DrudeForce below
                continue
            _found = False
            ats = ff.get_eqt_for_bond(bond)
            _term = BondTerm(*ats, 0)
            try:
                bterm = ff.get_bond_term(*ats)
                _found = True
            except:
                if transfer_bonded_terms:
                    bterm, score = dff_fuzzy_match(_term, ff)
                    if bterm is not None:
                        _found = True
                        _bterm_transferred.add((_term.name, bterm.name, score))
            if not _found:
                _bterm_not_found[_term.name] = _term
            else:
                self._ff.add_term(bterm, replace=True)
                self.bond_terms[bond] = bterm
                if bterm.fixed:
                    self.constrain_bonds[bond] = bterm.length

        _aterm_not_found = {}
        _aterm_transferred = set()
        for angle in self._topology.angles:
            _found = False
            ats_list = ff.get_eqt_for_angle(angle)
            _term = AngleTerm(*ats_list[0], 0)
            for ats in ats_list:
                try:
                    aterm = ff.get_angle_term(*ats)
                except FFTermNotFoundError:
                    pass
                else:
                    _found = True
                    break
            else:
                if transfer_bonded_terms:
                    aterm, score = dff_fuzzy_match(_term, ff)
                    if aterm is not None:
                        _found = True
                        _aterm_transferred.add((_term.name, aterm.name, score))
            if not _found:
                _aterm_not_found[_term.name] = _term
            else:
                self._ff.add_term(aterm, replace=True)
                self.angle_terms[angle] = aterm
                if not aterm.fixed:
                    continue
                _bond1 = Bond(angle.atom1, angle.atom2)
                _bond2 = Bond(angle.atom2, angle.atom3)
                bond1 = next(b for b in angle.atom2.bonds if b.equals(_bond1))
                bond2 = next(b for b in angle.atom2.bonds if b.equals(_bond2))
                # constrain the angle only if two bonds are also constrained
                if bond1 in self.constrain_bonds and bond2 in self.constrain_bonds:
                    d1, d2 = self.constrain_bonds[bond1], self.constrain_bonds[bond2]
                    self.constrain_angles[angle] = math.sqrt(d1 * d1 + d2 * d2 - 2 * d1 * d2 * math.cos(aterm.theta))

        _dterm_not_found = {}
        _dterm_transferred = set()
        for mol in self._topology.molecules:
            for dihedral in mol.dihedrals:
                _found = False
                ats_list = ff.get_eqt_for_dihedral(dihedral)
                _term = DihedralTerm(*ats_list[0])
                for ats in ats_list:
                    try:
                        dterm = ff.get_dihedral_term(*ats)
                    except FFTermNotFoundError:
                        pass
                    else:
                        _found = True
                        break
                else:
                    if transfer_bonded_terms:
                        dterm, score = dff_fuzzy_match(_term, ff)
                        if dterm is not None:
                            _found = True
                            _dterm_transferred.add((_term.name, dterm.name, score))
                if not _found:
                    _dterm_not_found[_term.name] = _term
                else:
                    self._ff.add_term(dterm, replace=True)
                    self.dihedral_terms[dihedral] = dterm

        _iterm_not_found = {}
        _iterm_transferred = set()
        for improper in self._topology.impropers:
            _found = False
            ats_list = ff.get_eqt_for_improper(improper)
            _term = ImproperTerm(*ats_list[0])
            for ats in ats_list:
                try:
                    iterm = ff.get_improper_term(*ats)
                except FFTermNotFoundError:
                    pass
                else:
                    _found = True
                    break
            else:
                if transfer_bonded_terms:
                    iterm, score = dff_fuzzy_match(_term, ff)
                    if iterm is not None:
                        _found = True
                        _iterm_transferred.add((_term.name, iterm.name, score))
            if not _found:
                if ignore_missing_improper:
                    iterm = OplsImproperTerm(*ats_list[0], 0.0)
                    _found = True
                else:
                    _iterm_not_found[_term.name] = _term
            if _found:
                self._ff.add_term(iterm, replace=True)
                self.improper_terms[improper] = iterm

        _pterm_not_found = {}
        for parent in self.drude_pairs.keys():
            _term = PolarizableTerm(ff.atom_types[parent.type].eqt_polar)
            try:
                pterm = ff.polarizable_terms[_term.name]
            except:
                _pterm_not_found[_term.name] = _term
            else:
                self._ff.add_term(pterm, replace=True)
                self.polarizable_terms[parent] = pterm

        ### print all transferred terms
        _n_transferred = 0
        msg = ''
        for k, v in {'Bond'    : _bterm_transferred,
                     'Angle'   : _aterm_transferred,
                     'Dihedral': _dterm_transferred,
                     'Improper': _iterm_transferred}.items():
            for i in v:
                _n_transferred += 1
                msg += '        %-10s %-20s %-20s %i\n' % (k, *i)
        if _n_transferred > 0:
            logger.warning(
                '%i bonded terms not found in FF. Transferred from similar terms with score\n' % _n_transferred + msg)

        ### print all the missing terms
        for d in _vdw_not_found, _bterm_not_found, _aterm_not_found, _dterm_not_found, _iterm_not_found, _pterm_not_found:
            self.missing_terms.extend(d.values())

        if self.missing_terms:
            msg = f'{len(self.missing_terms)} terms not found in FF'
            for term in self.missing_terms:
                msg += f'\n        {term}'
            logger.error(msg)
            if not allow_missing_terms:
                raise Exception('Incomplete force field')

        self.vdw_classes = {term.__class__ for term in self._ff.vdw_terms.values()}
        self.vdw_classes.update({term.__class__ for term in self._ff.pairwise_vdw_terms.values()})
        self.bond_classes = {term.__class__ for term in self.bond_terms.values()}
        self.angle_classes = {term.__class__ for term in self.angle_terms.values()}
        self.dihedral_classes = {term.__class__ for term in self.dihedral_terms.values()}
        self.improper_classes = {term.__class__ for term in self.improper_terms.values()}
        self.polarizable_classes = {term.__class__ for term in self.polarizable_terms.values()}
        self.ff_classes = (self.vdw_classes
                           .union(self.bond_classes)
                           .union(self.angle_classes)
                           .union(self.dihedral_classes)
                           .union(self.improper_classes)
                           .union(self.polarizable_classes))

    def export_lammps(self, data_out='data.lmp', in_out='in.lmp', **kwargs):
        '''
        Generate input files for Lammps.

        Parameters
        ----------
        data_out : str
        in_out : str
        '''
        from .lmpexporter import LammpsExporter
        LammpsExporter.export(self, data_out, in_out, **kwargs)

    def export_gromacs(self, gro_out='conf.gro', top_out='topol.top', mdp_out='grompp.mdp', **kwargs):
        '''
        Generate input files for GROMACS

        Parameters
        ----------
        gro_out : str or None
        top_out : str or None
        mdp_out : str or None
        '''
        from .gmxexporter import GromacsExporter
        GromacsExporter.export(self, gro_out, top_out, mdp_out, **kwargs)

    def export_namd(self, pdb_out='conf.pdb', psf_out='top.psf', prm_out='ff.prm', **kwargs):
        '''
        Generate input files for NAMD

        Parameters
        ----------
        pdb_out : str or None
        psf_out : str or None
        prm_out : str or None
        '''
        from .namdexporter import NamdExporter
        NamdExporter.export(self, pdb_out, psf_out, prm_out, **kwargs)

    def to_omm_system(self, disable_inter_mol=False, **kwargs):
        '''
        Export this system to OpenMM system

        Parameters
        ----------
        disable_inter_mol : bool

        Returns
        -------
        omm_system : openmm.openmm.System
        '''
        from .ommexporter import OpenMMExporter
        return OpenMMExporter.export(self, disable_inter_mol=disable_inter_mol, **kwargs)

    def decompose_energy(self, disable_inter_mol=False, verbose=True):
        '''
        Calculate the contribution of each energy term in this system

        Parameters
        ----------
        disable_inter_mol : bool
        verbose : bool

        Returns
        -------
        energies : Dict[str, float]
            The potential energies of different contributions
        '''
        try:
            from openmm import openmm as mm, app
            from mstk.ommhelper.utils import energy_decomposition
        except ImportError:
            raise Exception('OpenMM is required for energy decomposition')

        omm_sys = self.to_omm_system(disable_inter_mol=disable_inter_mol)
        integrator = mm.VerletIntegrator(0.001)
        platform = mm.Platform.getPlatformByName('CPU')
        sim = app.Simulation(self.topology.to_omm_topology(), omm_sys, integrator, platform)
        sim.context.setPositions(self.topology.positions)
        return energy_decomposition(sim, logger=logger if verbose else None)

    def minimize_energy(self, disable_inter_mol=False, tolerance=100.0, verbose=True):
        '''
        Perform energy minimization on this system.

        The position of each atom in the topology will be updated.

        Parameters
        ----------
        disable_inter_mol : bool
        tolerance : float
        verbose : bool
        '''
        try:
            from openmm import openmm as mm, app
            from mstk.ommhelper.utils import minimize
        except ImportError:
            raise Exception('OpenMM is required for energy minimization')

        omm_sys = self.to_omm_system(disable_inter_mol=disable_inter_mol)
        integrator = mm.VerletIntegrator(0.001)
        platform = mm.Platform.getPlatformByName('CPU')
        sim = app.Simulation(self.topology.to_omm_topology(), omm_sys, integrator, platform)
        sim.context.setPositions(self.topology.positions)
        minimize(sim, tolerance=tolerance, logger=logger if verbose else None)
        positions = sim.context.getState(getPositions=True).getPositions(asNumpy=True)._value
        self.topology.set_positions(positions)

    def get_TIP4P_linear_coeffs(self, atom):
        '''
        Get the three linear coefficients to calculate the position of TIP4P virtual site from O, H and H positions

        Parameters
        ----------
        atom : Atom
            The virtual site atom in TIP4P water topology

        Returns
        -------
        params : tuple of float
        '''
        vsite: TIP4PSite = atom.virtual_site
        if type(vsite) is not TIP4PSite:
            raise Exception(f'{atom} is not a TIP4PSite')

        O, H1, H2 = vsite.parents
        mol = atom.molecule
        _bond_OH = Bond(O, H1)
        try:
            bond = next(b for b in mol.bonds if b.equals(_bond_OH))
        except StopIteration:
            raise Exception(f'Corresponding O-H bond not found for virtual atom {atom}')
        bterm = self.bond_terms[bond]

        _angle_HOH = Angle(H1, O, H2)
        try:
            angle = next(a for a in mol.angles if a.equals(_angle_HOH))
        except StopIteration:
            raise Exception(f'Corresponding H-O-H angle not found for virtual atom {atom}')
        aterm = self.angle_terms[angle]

        offset = atom.virtual_site.parameters[0]
        c = offset / bterm.length / 2 / math.cos(aterm.theta / 2)

        return 1 - 2 * c, c, c



from mstk.topology import Bond
from .zft import ZftTyper


class GaffTyper(ZftTyper):
    '''
    `GaffTyper` is specifically designed for GAFF atom type assignment.
    It should only be used with the type definition file `gaff.zft`

    Compared with `ZftTyper`, this class will handle conjugated atom types according to the GAFF convention,
    e.g. cc/cd, ce/cf etc...

    The official tool for GAFF - antechamber has plenty of issues:
    - Mysterious aromatic assignment
    - A lost of parameters missing. `parmchk2` gives unreasonable guess for torsion parameters, e.g. biphenyl

    Parameters
    ----------
    file : str or file-like object, optional
        Type definition file.
        If the file does not exist, will search it under directories defined by `MSTK_FORCEFIELD_PATH`.

    Notes
    -----
    * SMARTS is parsed by using RDKit package. Make sure it is installed.
    * In type definition file, empty lines are ignored, and comments should start with ##.

    '''

    def __init__(self, file=None):
        super().__init__(file)

    def _type_molecule(self, molecule):
        '''
        Assign atom types in the molecule with rules in type definition file.

        The :attr:`~mstk.topology.Atom.type` attribute of all atoms in the molecule will be updated.

        ZftTyper use RDKit to do SMARTS matching, therefore the orders must be set for all the bonds in the molecule.
        Usually it means the molecule be initialized from SMILES.
        with :func:`~mstk.topology.Molecule.from_smiles` or :func:`~mstk.topology.Molecule.from_pybel`

        If an atom can not match any type by the predefined SMARTS patterns, an Exception will be raised.

        Parameters
        ----------
        molecule : Molecule
        '''
        super()._type_molecule(molecule)

        # find the chain of conjugated atom types, add chose one from the two types for each atom
        conj_atoms = [atom for atom in molecule.atoms if '|' in atom.type]
        if not conj_atoms:
            return

        conj1_partners = {atom: [] for atom in conj_atoms}
        conj2_partners = {atom: [] for atom in conj_atoms}
        for atom in conj_atoms:
            for bond in atom.bonds:
                partner = bond.atom2 if bond.atom1 is atom else bond.atom1
                if partner not in conj_atoms:
                    continue
                if bond.order == Bond.Order.DOUBLE or bond.is_aromatic:
                    conj2_partners[atom].append(partner)
                else:
                    conj1_partners[atom].append(partner)

        flag = {atom: False for atom in conj_atoms}

        def set_conj_type_partners(atom):
            type_suffix = atom.type[-1]
            for partner in conj1_partners[atom]:
                if flag[partner]:
                    continue
                candidates = partner.type.split('|')
                try:
                    partner.type = next(typ for typ in candidates if typ.endswith(type_suffix))
                except StopIteration:
                    partner.type = candidates[0]
                flag[partner] = True
                set_conj_type_partners(partner)
            for partner in conj2_partners[atom]:
                if flag[partner]:
                    continue
                candidates = partner.type.split('|')
                try:
                    partner.type = next(typ for typ in candidates if not typ.endswith(type_suffix))
                except StopIteration:
                    partner.type = candidates[0]
                flag[partner] = True
                set_conj_type_partners(partner)

        for atom in conj_atoms:
            if flag[atom]:
                continue
            atom.type = atom.type.split('|')[0]
            flag[atom] = True
            set_conj_type_partners(atom)



import itertools
import numpy as np


class NeighborList:
    '''
    Parameters
    ----------
    box : np.ndarray of shape (3,)
    cutoff : float

    Attributes
    ----------
    box : np.ndarray of shape (3,)
    n_cell : np.ndarray of shape (3,)
    cell_size : np.ndarray of shape (3,)
    cell_indexes : list of tuple of int
    '''

    def __init__(self, box, cutoff):
        if any(box / 3 < cutoff):
            raise Exception('Cutoff larger than box/3')
        n = np.zeros(3, dtype=int)
        for i in range(3):
            while True:
                if box[i] / (n[i] + 1) < cutoff:
                    break
                n[i] += 1

        self.box = box
        self.n_cell = n
        self.cell_indexes = list(itertools.product(range(n[0]), range(n[1]), range(n[2])))
        self.cell_size = box / n
        self._cells = [[[[] for _ in range(n[2])] for _ in range(n[1])] for _ in range(n[0])]

    def build(self, positions):
        '''
        Parameters
        ----------
        positions : np.ndarray
        '''
        positions -= np.floor(positions / self.box) * self.box
        locations = np.clip(np.floor(positions / self.cell_size).astype(int), 0, self.n_cell - 1)
        for i, (x, y, z) in enumerate(locations):
            self._cells[x][y][z].append(i)

    def get_cell(self, index):
        '''
        Parameters
        ----------
        index : tuple of int

        Returns
        -------
        index_atoms : list of int
        '''
        return self._cells[index[0]][index[1]][index[2]]

    def get_interacting_cell(self, index):
        '''
        Return the 27 interacting cells as a single large cell

        Parameters
        ----------
        index : tuple of int

        Returns
        -------
        index_atoms : list of int
        '''
        idx_surround = [[], [], []]
        for i in range(3):
            idx = index[i]
            idx_surround[i] = [idx - 1 if idx - 1 >= 0 else self.n_cell[i] - 1,
                               idx,
                               idx + 1 if idx + 1 <= self.n_cell[i] - 1 else 0
                               ]
        cell = []
        for index in itertools.product(*idx_surround):
            cell.extend(self.get_cell(index))

        return cell



#!/usr/bin/env python3

import sys
import pickle
import traceback
import argparse
import numpy as np
from openmm import openmm as mm
from mstk.topology import Topology, Molecule
from mstk.trajectory import Trajectory
from mstk.forcefield import ForceField
from mstk.simsys import System
from mstk.chem import constant
from mstk import logger


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--top', type=str, required=True,
                        help='topology file for the system. '
                             'This topology must have charges and masses assigned. '
                             'The charge and mass from force field will not be used')
    parser.add_argument('-f', '--ff', nargs='+', type=str, required=True, help='forcefield files')
    parser.add_argument('-c', '--conf', type=str, required=True,
                        help='configuration file with positions and box. The last frame will be used')
    parser.add_argument('--box', nargs='+', type=float,
                        help='periodic box size. '
                             'The box will be rectangular if three values are provided. '
                             'The box will be cubic if only one value is provided.'
                        )
    parser.add_argument('--density', type=float,
                        help='the density of the target box in g/mL. '
                             'The size of cubic box will be calculated from it. '
                             'If the argument box is provided, density will be ignored')
    parser.add_argument('--lammps', action='store_true', help='export LAMMPS files')
    parser.add_argument('--gromacs', action='store_true', help='export GROMACS files')
    parser.add_argument('--namd', action='store_true', help='export NAMD files')
    parser.add_argument('--openmm', action='store_true', help='export OpenMM files')
    parser.add_argument('--all', action='store_true', help='export files for all supported MD engines')
    return parser.parse_args()


def main():
    args = parse_args()
    for arg, val in vars(args).items():
        logger.info(f'--{arg:10s} {val}')

    top = Topology.open(args.top)
    ff = ForceField.open(*args.ff)

    frame = Trajectory.read_frame_from_file(args.conf, -1)
    if frame.cell.volume != 0:
        top.cell.set_box(frame.cell.vectors)

    # the frame may not contain Drude particles and virtual sites
    _positions_set = False
    if len(frame.positions) == top.n_atom:
        top.set_positions(frame.positions)
        _positions_set = True
    elif ff.is_polarizable or ff.has_virtual_site:
        atoms_real = [atom for atom in top.atoms if not atom.is_drude and atom.virtual_site is None]
        if len(frame.positions) == len(atoms_real):
            for i, atom in enumerate(atoms_real):
                atom.position = frame.positions[i][:]
            np.random.seed(1)
            for parent, drude in top.get_drude_pairs():
                drude.position = parent.position + (np.random.random(3) - 0.5) / 100
            for parent, vsite in top.get_virtual_site_pairs():
                vsite.position = vsite.virtual_site.calc_position()
            _positions_set = True

    if not _positions_set:
        logger.error('Numbers of atoms in trajectory (%i) and topology (all: %i, real: %i) do not match' % (
            len(frame.positions), top.n_atom,
            top.n_atom - len(top.get_drude_pairs()) - len(top.get_virtual_site_pairs())))
        sys.exit(1)

    box = [0.0, 0.0, 0.0]
    if args.density is not None:
        mass = sum(atom.mass for atom in top.atoms)  # g/mol
        vol = mass / args.density / constant.AVOGADRO  # cm^3
        box = [vol ** (1 / 3) * constant.CENTI / constant.NANO] * 3  # nm

    if args.box is not None:
        if len(args.box) == 3:
            box = args.box
        elif len(args.box) == 1:
            box = args.box * 3
        else:
            raise Exception('box should be one or three floats')

    logger.info(f'box = {box}')
    top.cell.set_box(box)

    system = System(top, ff)

    if args.lammps or args.all:
        logger.info('Exporting LAMMPS...')
        try:
            system.export_lammps(data_out='_lmp-data.lmp', in_out='_lmp-in.lmp')
        except Exception as e:
            logger.error('Failed exporting LAMMPS')
            traceback.print_exc()

    if args.gromacs or args.all:
        logger.info('Exporting GROMACS...')
        try:
            system.export_gromacs(gro_out='_gmx-conf.gro', top_out='_gmx-topol.top', mdp_out='_gmx-pp.mdp')
        except Exception as e:
            logger.error('Failed exporting GROMACS')
            traceback.print_exc()

    if args.namd or args.all:
        logger.info('Exporting NAMD...')
        try:
            system.export_namd(pdb_out='_namd-conf.pdb', psf_out='_namd-top.psf', prm_out=None)
        except Exception as e:
            logger.error('Failed exporting NAMD')
            traceback.print_exc()

    if args.openmm or args.all:
        logger.info('Exporting OpenMM...')
        try:
            omm_top = top.to_omm_topology()
            with open('_omm-top.pkl', 'wb') as f:
                f.write(pickle.dumps(omm_top))
            omm_sys = system.to_omm_system()
            with open('_omm-sys.xml', 'w') as f:
                f.write(mm.XmlSerializer.serialize(omm_sys))
        except Exception as e:
            logger.error('Failed exporting OpenMM')
            traceback.print_exc()


if __name__ == '__main__':
    main()



from mstk.topology.molecule import Molecule
from mstk.topology.topology import Topology


class Smi():
    '''
    Generate Topology from SMI file.

    Each line is a molecule represented by SMILES string.
    Molecule name can optionally be put after the SMILES string, separated by spaces.
    Therefore, the moelcule name itself cannot contain space.
    Lines started with # will be treated as comments.
    The positions will be generated using RDKit.

    Parameters
    ----------
    file : str
    kwargs : dict
        Ignored

    Attributes
    ----------
    topology : Topology

    Examples
    --------
    >>> smi = Smi('input.zmat')
    >>> topology = smi.topology
    '''

    def __init__(self, file, **kwargs):
        self.topology = Topology()
        self._parse(file)

    def _parse(self, file):
        molecules = []
        with open(file) as f:
            for line in f:
                line = line.strip()
                if line == '' or line.startswith('#'):
                    continue
                mol = Molecule.from_smiles(line)
                molecules.append(mol)

        self.topology.update_molecules(molecules)


Topology.register_format('.smi', Smi)



#!/usr/bin/env python3

import argparse
import os
import re
import pandas as pd
from mstk.analyzer.series import block_average, is_converged
from mstk.analyzer.fitting import polyfit, polyval
from mstk.utils import print_data_to_file


def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('input', type=str, help='log file of OpenMM or LAMMPS')
    parser.add_argument('-b', '--begin', default=-1, type=float, help='begin from this data')
    parser.add_argument('-e', '--end', default=-1, type=float, help='end at this data')
    parser.add_argument('--skip', default=1, type=int, help='read every N rows')
    parser.add_argument('-o', '--output', type=str, help='output data to a file')
    parser.add_argument('-c', '--converge', default=False, action='store_true', help='detect convergence')
    parser.add_argument('-p', '--plot', default=False, action='store_true', help='plot data')
    parser.add_argument('--type', type=str, choices=['openmm', 'lammps', 'xvg'])
    parser.add_argument('--fit', action='store_true', help='perform linear fitting')
    parser.add_argument('--dist', action='store_true', help='plot distribution')

    return parser.parse_args()


class Analyzer:
    def __init__(self, log_file, file_type=None, begin=-1, end=-1, skip=1):
        self.labels: [str] = []
        self.data_list: [[float]] = []
        self.when_list: [int] = []
        self.fit_coeff: [(float)] = []
        self.log_file = log_file

        _ext_filetype = {
            'xvg': 'xvg',
            'out': 'openmm',
            'txt': 'openmm',
            'log': 'lammps'
        }
        ext = os.path.splitext(log_file)[1].lstrip('.').lower()
        file_type = file_type or _ext_filetype.get(ext, ext)
        if file_type == 'xvg':
            self.read_xvg(begin, end, skip)
        elif file_type == 'openmm':
            self.read_openmm_log(begin, end, skip)
        elif file_type == 'lammps':
            self.read_lammps_log(begin, end, skip)
        else:
            raise Exception('unknown log file type: ' + file_type)

        # in case times in all frames are zero
        if all(i == 0 for i in self.data_list[0]):
            self.data_list[0] = [i for i in range(len(self.data_list[0]))]

    @staticmethod
    def str2float(string):
        try:
            val = float(string)
        except ValueError:
            val = 0
        return val

    def read_openmm_log(self, begin, end, skip):
        n_row = -1
        _START = False
        for line in open(self.log_file):
            if line.startswith('#"'):
                if not _START:
                    self.labels = [s.strip('"') for s in line.strip('#').strip().split('\t')]
                    self.data_list = [[] for _ in self.labels]
                    self.when_list = [0 for _ in self.labels]
                    _START = True
                continue
            if _START:
                words = line.strip().split()
                if len(words) != len(self.labels):
                    continue
                try:
                    values = list(map(Analyzer.str2float, words))
                except ValueError:
                    continue
                step = values[0]
                if step < begin:
                    continue
                if 0 < end < step:
                    break
                n_row += 1
                if n_row % skip != 0:
                    continue
                for i in range(len(self.labels)):
                    self.data_list[i].append(values[i])

    def read_lammps_log(self, begin, end, skip):
        with open(self.log_file) as f:
            content = f.read()
        nMin = len(re.findall('\nMinimization', content))
        nRun = len(re.findall('\nStep', content)) - nMin
        print('%i Min, %i Run' % (nMin, nRun))

        n_row = -1
        _START = False
        i_cycle = -nMin
        for line in content.splitlines():
            if line.startswith('Step'):
                i_cycle += 1
                if i_cycle > 0:
                    _START = True
                if i_cycle == 1:
                    self.labels = line.strip().split()
                    self.data_list = [[] for _ in self.labels]
                    self.when_list = [0 for _ in self.labels]
                continue
            if i_cycle > 0 and line.startswith('Loop'):
                _START = False
                continue
            if i_cycle > 0 and _START:
                try:
                    step = int(line.strip().split()[0])
                except:
                    continue
                if step < begin:
                    continue
                if 0 < end < step:
                    break
                n_row += 1
                if n_row % skip != 0:
                    continue
                for i in range(len(self.labels)):
                    self.data_list[i].append(float(line.strip().split()[i]))

    def read_xvg(self, begin, end, skip):
        n_row = -1
        for line in open(self.log_file):
            if line.startswith('#'):
                continue
            if line.startswith('@'):
                words = line.strip().split()
                if words[1] in ('xaxis', 'yaxis'):
                    self.labels.append(words[3].strip('"').split()[0])
                    self.data_list.append([])
                    self.when_list.append(0)
                continue
            words = line.strip().split()
            step = float(words[0])
            if step < begin:
                continue
            if 0 < end < step:
                break
            n_row += 1
            if n_row % skip != 0:
                continue
            for i in range(len(self.data_list)):
                self.data_list[i].append(float(words[i]))

    def detect_converge(self):
        for i in range(1, len(self.data_list)):
            series = pd.Series(self.data_list[i], index=list(range(len(self.data_list[0]))))
            converged, when = is_converged(series, frac_min=0)
            self.when_list[i] = when

    def fit(self):
        self.fit_coeff.append((0.0, 1.0))
        for i in range(1, len(self.data_list)):
            when = self.when_list[i]
            coeff, score = polyfit(self.data_list[0][when:], self.data_list[i][when:], degree=1)
            self.fit_coeff.append(coeff)

    def print_summary(self):
        string = 'File: %s, Steps: %i-%i, Samples: %i\n' % (
            self.log_file, self.data_list[0][0], self.data_list[0][-1], len(self.data_list[0]))
        string += ' %2s %12s %11s %8s %8s %8s %9s %9s\n' % (
            'ID', 'LABEL', 'MEAN', 'STDERR', 'STDEV', 'WHEN', 'INTERCEPT', 'SLOPE')
        for i in range(1, len(self.labels)):
            when = self.when_list[i]
            data = self.data_list[i][when:]
            if self.fit_coeff:
                intercept, slope = self.fit_coeff[i]
            else:
                intercept, slope = 0, 0
            (ave, err_ave), (std, err_std) = block_average(data)
            string += ' %2i %12s %11.5g %8.2g %8.2g %8.4g %9.3g %9.3g\n' % (
                i, self.labels[i], ave, err_ave, std, self.data_list[0][when], intercept, slope)

        print(string, end='')

    def plot_data(self, plot_distribution=False):
        import matplotlib
        import matplotlib.pyplot as plt

        matplotlib.use('Agg')
        matplotlib.rcParams.update({'font.size': 15})

        for idx in range(1, len(self.data_list)):
            when = self.when_list[idx]
            if not plot_distribution:
                fig, ax1 = plt.subplots()
            else:
                fig, (ax1, ax2) = plt.subplots(1, 2)
                ax2.hist(self.data_list[idx][when:], density=True, bins=40, color='C1')
                ax2.set_xlabel(self.labels[idx])
                ax2.set_ylabel('Probability')
            ax1.plot(self.data_list[0][:when], self.data_list[idx][:when])
            x_list, y_list = self.data_list[0][when:], self.data_list[idx][when:]
            ax1.plot(x_list, y_list)
            if self.fit_coeff:
                y_pred_list = [polyval(x, self.fit_coeff[idx]) for x in x_list]
                ax1.plot(x_list, y_pred_list, '--')
            ax1.set_xlabel(self.labels[0])
            ax1.set_ylabel(self.labels[idx])
            fig.tight_layout()
            plt.savefig(self.labels[idx] + '.png')
            plt.show()


if __name__ == '__main__':
    args = parse_args()

    analyzer = Analyzer(args.input, args.type, args.begin, args.end, args.skip)
    if args.converge:
        analyzer.detect_converge()
    if args.fit:
        analyzer.fit()
    analyzer.print_summary()
    if args.plot:
        analyzer.plot_data(plot_distribution=args.dist)
    if args.output:
        print_data_to_file(dict(zip(analyzer.labels, analyzer.data_list)), file=args.output)



import os
from mstk.utils import random_string


class Gauss:
    '''
    Wrapper for Gaussian for quantum calculations

    Parameters
    ----------
    gauss_bin : str
        Path of the Gaussian binary executable
    scrdir : str, optional
        Path of the scratch dir for Gaussian calculations.
        If not provided, the ramdisk `/dev/shm` will be used.
    '''

    _TEMPLATE_DIR = os.path.abspath(os.path.dirname(__file__) + os.sep + '../template/gauss/')

    def __init__(self, gauss_bin, scrdir=None):
        if scrdir is None:
            scrdir = '/dev/shm'
        self.GAUSS_BIN = gauss_bin
        self.GAUSS_ROOT = os.path.dirname(gauss_bin)
        self.SCRDIR = scrdir

    def generate_gjf_cv(self, name, py_mol, method='B3LYP', basis='6-31G*', scale=0.9613,
                        T_list=None):
        '''
        Genrate GJF input file for calculating heat capacity with Gaussian

        The molecule is provided through a pybel.Molecule object.
        Therefore, openbabel should be installed.
        The pybel.Molecule object should contain information about positions of all atoms.

        The hindered-rotor model is used to describing the low-frequency dihedral vibrations.

        A list of temperature can be provided.
        Then the thermodynamic analysis will be performed at all the provided temperatures,
        so that the result at intermediate temperature can be interpolated.

        Parameters
        ----------
        name : str
        py_mol : pybel.Molecule
        method : str
        basis : str
        scale : float
        T_list : list of float, optional
        '''
        if T_list is None:
            T_list = [298]
        gjf = name + '.gjf'
        with open(gjf, 'w') as f:
            f.write('%%chk=%(name)s.chk\n'
                    '# opt freq=hindrot %(method)s %(basis)s scale=%(scale).4f temperature=%(T).2f\n'
                    '\n'
                    'Title\n'
                    '\n'
                    '%(charge)i %(multiplicity)i\n'
                    % ({'name'        : name,
                        'method'      : method,
                        'basis'       : basis,
                        'scale'       : scale,
                        'T'           : T_list[0],
                        'charge'      : py_mol.charge,
                        'multiplicity': py_mol.spin
                        })
                    )
            for atom_line in py_mol.write('xyz').splitlines()[2:]:
                f.write(atom_line + '\n')
            f.write('\n')

            for T in T_list[1:]:
                f.write('--Link1--\n'
                        '%%chk=%(name)s.chk\n'
                        '# freq=(readfc,hindrot) geom=allcheck scale=%(scale).4f temperature=%(T).2f\n'
                        '\n'
                        % ({'name' : name,
                            'scale': scale,
                            'T'    : T
                            })
                        )

    def run_gjf(self, gjf, nprocs=1, memMB=1000, get_cmd=False, sh_out=None):
        '''
        Process the GJF file so that it is ready for multi-core Gaussian calculation, and then return the commands for calculation.

        The commands return can be feed into a :class:`~mstk.scheduler.Scheduler` for running on a HPC.

        Parameters
        ----------
        gjf : str
            The GJF file to be processed
        nprocs : str
            Number of cores to use
        memMB : int
            Maxium size of memory to use in unit of MB
        get_cmd : bool
            Whether or not return the commands for running Gaussian calculation
        sh_out : str, optional
            If provided, the commmands for running Gaussian calculation will be written to this shell script

        Returns
        -------
        commands : list of str, or None
            If `get_cmd` is True, the commands for running Gaussian will be returned
        '''
        with open(gjf) as f:
            content = f.read()
        with open(gjf, 'w') as f:
            f.write('%%nprocshared=%i\n' % nprocs)
            f.write('%%mem=%iMB\n' % memMB)
            f.write(content)

        JOB_DIR = os.path.join(self.SCRDIR, random_string(6))
        cmds = []
        cmds.append('JOB_DIR=' + JOB_DIR)
        cmds.append('mkdir -p ${JOB_DIR}')
        cmds.append('export GAUSS_EXEDIR=%s:%s/bsd' % (self.GAUSS_ROOT, self.GAUSS_ROOT))
        cmds.append('export GAUSS_SCRDIR=${JOB_DIR}')
        cmds.append('%s %s' % (self.GAUSS_BIN, gjf))
        cmds.append('rm -rf ${JOB_DIR}')

        if sh_out != None:
            with open(sh_out, 'w') as f:
                f.write('\n'.join(cmds))
        if get_cmd:
            return cmds



import numpy as np


def polyfit(x, y, degree, weight=None):
    '''
    Least square n-th order polynomial fitting:
    y = c0 + c1 * x + c2 * x**2 + ... + cn * x**n

    Parameters
    ----------
    x : list of float
    y : list of float
    degree : int
    weight : list of float, optional

    Returns
    -------
    coeff : ndarray
    rsq : float
    '''
    from sklearn import linear_model
    from sklearn.preprocessing import PolynomialFeatures

    skx = list(zip(x))
    skv = list(y)

    skx = np.array(skx).astype(np.float64)
    skv = np.array(skv).astype(np.float64)

    poly = PolynomialFeatures(degree)
    skx_ = poly.fit_transform(skx)
    clf = linear_model.LinearRegression(fit_intercept=False)
    clf.fit(skx_, skv, sample_weight=weight)
    return clf.coef_, clf.score(skx_, skv)


def polyval(x, coeff):
    '''
    Evaluate the n-th order polynomial result:
    y = c0 + c1 * x + c2 * x**2 + ... + cn * x**n

    Parameters
    ----------
    x : float
    coeff : list of float

    Returns
    -------
    y : float
    '''
    from numpy.polynomial.polynomial import polyval as np_polyval
    return np_polyval(x, coeff)


def polyval_derivative(x, coeff):
    '''
    Evaluate the n-th order polynomial result and derivative:
    y = c0 + c1 * x + c2 * x**2 + ... + cn * x**n;
    dy/dx = c1 + 2*c2 * x + ... + n*cn * x**(n-1)

    Parameters
    ----------
    x : float
    coeff : list of float

    Returns
    -------
    y : float
    dy/dx : float
    '''
    from numpy.polynomial.polynomial import polyval as np_polyval
    y = np_polyval(x, coeff)

    degree = len(coeff) - 1
    dydx = 0
    for i in range(degree):
        dydx += (i + 1) * coeff[i + 1] * x ** i

    return y, dydx


def polyfit_2d(x, y, z, degree, weight=None):
    '''
    Least square n-th order 2-D polynomial fitting.
    e.g. the 3rd polynomial fitting:
    z = c0 + c1 * x + c2 * y + c3 * xx + c4 * xy + c5 * yy + c6 * xxx + c7 * xxy + c8 * xyy + c9 * yyy

    Parameters
    ----------
    x : list of float
    y : list of float
    z : list of float
    degree : int
    weight : list of float, optional

    Returns
    -------
    coeff : ndarray
    rsq : float
    '''
    from sklearn import linear_model
    from sklearn.preprocessing import PolynomialFeatures

    skx = list(zip(x, y))
    skv = list(z)

    skx = np.array(skx).astype(np.float64)
    skv = np.array(skv).astype(np.float64)

    poly = PolynomialFeatures(degree)
    skx_ = poly.fit_transform(skx)
    clf = linear_model.LinearRegression(fit_intercept=False)
    clf.fit(skx_, skv, sample_weight=weight)
    return clf.coef_, clf.score(skx_, skv)


def polyval_derivative_2d(x, y, degree, coeff):
    '''
    Evaluate the n-th order 2-D polynomial result and derivative. Only 3-rd and 4-th order are supported.
    e.g. the 3rd polynomial fitting:
    z = c0 + c1 * x + c2 * y + c3 * xx + c4 * xy + c5 * yy + c6 * xxx + c7 * xxy + c8 * xyy + c9 * yyy

    Parameters
    ----------
    x : float
    y : float
    degree : [3, 4]
    coeff : list of float

    Returns
    -------
    z : float
    dz/dx : float
    dz/dy : float
    '''
    if degree == 3:
        k0, k1, k2, k3, k4, k5, k6, k7, k8, k9 = coeff
        z = k0 \
            + k1 * x + k2 * y \
            + k3 * x * x + k4 * x * y + k5 * y * y \
            + k6 * x * x * x + k7 * x * x * y + k8 * x * y * y + k9 * y * y * y
        dzdx = k1 + 2 * k3 * x + k4 * y + 3 * k6 * x * x + 2 * k7 * y * x + k8 * y * y
        dzdy = k2 + k4 * x + 2 * k5 * y + k7 * x * x + 2 * k8 * x * y + 3 * k9 * y * y

    elif degree == 4:
        k0, k1, k2, k3, k4, k5, k6, k7, k8, k9, k10, k11, k12, k13, k14 = coeff
        z = k0 \
            + k1 * x + k2 * y \
            + k3 * x * x + k4 * x * y + k5 * y * y \
            + k6 * x * x * x + k7 * x * x * y + k8 * x * y * y + k9 * y * y * y \
            + k10 * x ** 4 + k11 * x ** 3 * y + k12 * x ** 2 * y ** 2 + k13 * x * y ** 3 + k14 * y ** 4
        dzdx = k1 + 2 * k3 * x + k4 * y + 3 * k6 * x * x + 2 * k7 * y * x + k8 * y * y \
               + 4 * k10 * x ** 3 + 3 * k11 * y * x ** 2 + 2 * k12 * y ** 2 * x + k13 * y ** 3
        dzdy = k2 + k4 * x + 2 * k5 * y + k7 * x * x + 2 * k8 * x * y + 3 * k9 * y * y \
               + k11 * x ** 3 + 2 * k12 * x ** 2 * y + 3 * k13 * x * y ** 2 + 4 * k14 * y ** 3

    else:
        raise Exception('Available degree: 3, 4')

    return z, dzdx, dzdy


def curve_fit_rsq(func, x_list, y_list, guess=None, bounds=None):
    '''
    Least square curve fitting

    Parameters
    ----------
    func : function
    x_list : list of float
    y_list : list of float
    guess : list of float, optional
    bounds : list of float, optional

    Returns
    -------
    coeff: tuple of float
    rsq : float
    '''
    from scipy.optimize import curve_fit

    x_array = np.array(x_list)
    y_array = np.array(y_list)

    popt, pcov = curve_fit(func, x_array, y_array, guess, bounds=bounds)
    ss_tot = ((y_array - y_array.mean()) ** 2).sum()
    predict = np.array([func(x, *popt) for x in x_array])
    ss_res = ((y_array - predict) ** 2).sum()
    rsq = 1 - ss_res / ss_tot

    return tuple(popt), rsq


def fit_opls_dihedral(x_list, y_list, guess=None, bounds=None):
    def func(x, k1, k2, k3):
        return k1 * (1 + np.cos(x)) + k2 * (1 - np.cos(2 * x)) + k3 * (1 + np.cos(3 * x))

    guess = guess or [0.0, 0.0, 0.0]
    bounds = bounds or (-np.inf, np.inf)

    return curve_fit_rsq(func, x_list, y_list, guess, bounds)


def tg_hyperbola(T, T0, d0, a, b, c):
    H = (T - T0) / 2 + np.sqrt((T - T0) ** 2 / 4 + np.exp(c))
    return d0 - a * (T - T0) - b * H


def fit_tg_hyperbola(T_list, d_list, guess=None, bounds=None):
    guess = guess or [400.0, 1.0, 0.0, 0.0, 0.0]
    bounds = bounds or ([0.0, 0.0, 0.0, 0.0, -np.inf], [1000.0, 2.0, np.inf, np.inf, np.inf])

    return curve_fit_rsq(tg_hyperbola, T_list, d_list, guess, bounds)


def logistic(x, A1, A2, x0, p):
    return (A1 - A2) / (1 + (x / x0) ** p) + A2


def logistic_derivative(x, A1, A2, x0, p):
    y = logistic(x, A1, A2, x0, p)
    dydx = -(A1 - A2) / (1 + (x / x0) ** p) ** 2 * p * (x / x0) ** p / x
    return y, dydx


def fit_logistic(x_list, y_list, guess=None, bounds=None):
    guess = guess or [y_list[0], 2 * y_list[-1] - y_list[0], x_list[-1], 1.0]
    bounds = bounds or ([-np.inf, -np.inf, -np.inf, 0], np.inf)

    return curve_fit_rsq(logistic, x_list, y_list, guess, bounds)


def fit_vle_tanh(x_list, d_list, guess=None, bounds=None):
    def func(x, c, A, r, s):
        return c + A * np.tanh((x - r) / s)

    guess = guess or [0, 1, 0, 1]
    bounds = bounds or (-np.inf, np.inf)

    return curve_fit_rsq(func, x_list, d_list, guess, bounds)


def vle_dminus(T, Tc, B):
    return B * (1 - T / Tc) ** 0.325


def fit_vle_dminus(T_list, dminus_list, guess=None, bounds=None):
    '''
    Fit critical temperature using VLE density:
    dliq - dgas = B(1-T/Tc)**0.325

    Parameters
    ----------
    T_list : list of float
    dminus_list : list of float
    guess : list of float, optional
    bounds : list of float, optional

    Returns
    -------
    coeff : tuple of float
        Tc and B
    rsq : float
    '''
    guess = guess or [max(T_list) / 0.8, 1.0]
    bounds = bounds or (0, np.inf)

    return curve_fit_rsq(vle_dminus, T_list, dminus_list, guess, bounds)


def vle_dplus(T, Dc, A, Tc):
    return 2 * (Dc + A * (1 - T / Tc))


def fit_vle_dplus(T_list, dplus_list, Tc, guess=None, bounds=None):
    '''
    Fit critical density using VLE density and critical temperature:
    dliq + dgas = 2(Dc+A(1-T/Tc))

    Parameters
    ----------
    T_list : list of float
    dplus_list : list of float
    Tc : float
    guess : list of float, optional
    bounds : list of float, optional

    Returns
    coeff : tuple of float
        Tc and B
    rsq : float
    -------
    '''
    guess = guess or [0.3, 1.0]
    bounds = bounds or (0, np.inf)

    return curve_fit_rsq(lambda T, Dc, A: vle_dplus(T, Dc, A, Tc), T_list, dplus_list, guess, bounds)


def vle_st(T, A, n, Tc):
    return A * (1 - T / Tc) ** n


def fit_vle_st(T_list, st_list, Tc, guess=None, bounds=None):
    guess = guess or [50, 1.22]
    bounds = bounds or (0, np.inf)

    return curve_fit_rsq(lambda T, A, n: vle_st(T, A, n, Tc), T_list, st_list, guess, bounds)


def vle_log10pvap(T, A, B):
    return A - B / T


def vle_pvap(T, A, B):
    return 10 ** vle_log10pvap(T, A, B)


def fit_vle_pvap(T_list, pvap_list, guess=None, bounds=None):
    y_array = np.log10(np.array(pvap_list))

    guess = guess or [10, 3000]
    bounds = bounds or (0, np.inf)

    return curve_fit_rsq(vle_log10pvap, T_list, y_array, guess, bounds)



'''
A set of commonly used small functions.
'''

import os
import subprocess
import random
import string
import numpy as np
from typing import Iterable


def greatest_common_divisor(numbers):
    '''
    Calculate the greatest common divisor.

    Parameters
    ----------
    numbers : list of int

    Returns
    divisor : int
    '''
    minimal = min(numbers)
    for i in range(minimal, 1, -1):
        flag = True
        for number in numbers:
            if number % i != 0:
                flag = False
        if flag:
            return i
    return 1


def random_string(length=8):
    '''
    Generate a random string in specified length. The string contains only upper and lower case ascii letters.


    Parameters
    ----------
    length : int

    Returns
    -------
    string : str
    '''
    return ''.join(random.sample(string.ascii_letters, length))


def cd_or_create_and_cd(dir):
    '''
    Go to the target directory. If not exist, create this directory and go to it.

    Parameters
    ----------
    dir : str
    '''
    if not os.path.exists(dir):
        try:
            os.makedirs(dir)
        except:
            raise Exception('Cannot create directory: %s' % dir)

    try:
        os.chdir(dir)
    except:
        raise Exception('Cannot read directory: %s' % dir)


def estimate_density_from_formula(f) -> float:
    # in unit of g/mL
    from mstk.chem.formula import Formula
    formula = Formula(f)
    string = formula.to_str()
    density = {
        'H2': 0.07,
        'He': 0.15,
    }
    if string in density.keys():
        return density.get(string)

    nAtoms = formula.n_heavy + formula.n_h
    nC = formula.atoms.get('C', 0)
    nH = formula.atoms.get('H', 0)
    nO = formula.atoms.get('O', 0)
    nN = formula.atoms.get('N', 0)
    nS = formula.atoms.get('S', 0)
    nF = formula.atoms.get('F', 0)
    nCl = formula.atoms.get('Cl', 0)
    nBr = formula.atoms.get('Br', 0)
    nI = formula.atoms.get('I', 0)
    nOther = nAtoms - nC - nH - nO - nN - nS - nF - nCl - nBr - nI
    return (1.175 * nC + 0.572 * nH + 1.774 * nO + 1.133 * nN + 2.184 * nS
            + 1.416 * nF + 2.199 * nCl + 5.558 * nBr + 7.460 * nI
            + 0.911 * nOther) / nAtoms


def n_diff_lines(f1: str, f2: str):
    with open(f1) as f:
        l1 = f.readlines()
    with open(f2) as f:
        l2 = f.readlines()

    n1 = len(l1)
    n2 = len(l2)

    n_diff = 0
    for i in range(min(n1, n2)):
        if l1[i].strip() != l2[i].strip():
            n_diff += 1
    n_diff += abs(n1 - n2)
    return n_diff


def get_last_line(filename, n=1):
    """Returns the nth before last line of a file (n=1 gives last line)"""
    num_newlines = 0
    with open(filename, 'rb') as f:
        try:
            f.seek(-2, os.SEEK_END)
            while num_newlines < n:
                f.seek(-2, os.SEEK_CUR)
                if f.read(1) == b'\n':
                    num_newlines += 1
        except OSError:
            f.seek(0)
        last_line = f.readline().decode()
    return last_line


def histogram(data, bins, normed=False):
    y, _x = np.histogram(data, bins=bins, density=normed)
    x = (_x[1:] + _x[:-1]) / 2
    return x, y


def print_data_to_file(name_column_dict, file):
    if len(name_column_dict) == 0:
        raise Exception('Columns are empty')

    with open(file, 'w') as f:
        f.write('#')
        for name in name_column_dict.keys():
            f.write(f'"{name}"\t')
        f.write('\n')
        for i in range(len(list(name_column_dict.values())[0])):
            for column in name_column_dict.values():
                f.write(f'{column[i]}\t')
            f.write('\n')


def align_mpl_axis(ax1, ax2):
    ylims1 = list(ax1.axes.get_ylim())
    ylims1[0] = min(0, ylims1[0])
    ylims1[1] = max(0, ylims1[1])
    ylims2 = list(ax2.axes.get_ylim())
    ylims2[0] = min(0, ylims2[0])
    ylims2[1] = max(0, ylims2[1])

    if ylims1[0] == 0:
        if ylims2[0] == 0:
            return
        elif ylims2[1] == 0:
            ylims1[0] = -ylims1[1]
            ylims2[1] = -ylims2[0]
        else:
            ylims1[0] = ylims1[1] / ylims2[1] * ylims2[0]
    elif ylims1[1] == 0:
        if ylims2[0] == 0:
            ylims1[1] = -ylims1[0]
            ylims2[0] = -ylims2[1]
        elif ylims2[1] == 0:
            return
        else:
            ylims1[1] = ylims1[0] / ylims2[0] * ylims2[1]
    else:
        if ylims2[0] == 0:
            ylims2[0] = ylims2[1] / ylims1[1] * ylims1[0]
        elif ylims2[1] == 0:
            ylims2[1] = ylims2[0] / ylims1[0] * ylims1[1]
        else:
            ratio1 = -ylims1[1] / ylims1[0]
            ratio2 = -ylims2[1] / ylims2[0]
            if ratio1 > ratio2:
                ylims2[1] = -ylims2[0] * ratio1
            else:
                ylims1[1] = -ylims1[0] * ratio2

    ax1.set(ylim=ylims1)
    ax2.set(ylim=ylims2)


def flatten(it):
    for x in it:
        if (isinstance(x, Iterable) and not isinstance(x, str)):
            yield from flatten(x)
        else:
            yield x



from .atom import Atom
from .virtualsite import *
from .connectivity import *
from .residue import Residue
from .molecule import  Molecule
from .topology import Topology
from .unitcell import UnitCell
from .io.psf import Psf
from .io.zmat import Zmat
from .io.pdb import Pdb
from .io.lammps import LammpsData
from .io.xyz import XyzTopology
from .io.msd import Msd
from .io.smi import Smi
from .io.gro import GroTopology
